import pytest

from dagster import (
    Any,
    Bool,
    DagsterInvalidConfigError,
    Float,
    InputDefinition,
    Int,
    List,
    Optional,
    OutputDefinition,
    PipelineDefinition,
    String,
    execute_pipeline,
    lambda_solid,
)
from dagster.utils.test import get_temp_file_name


def _execute_pipeline_with_subset(pipeline, run_config, solid_selection):
    return execute_pipeline(
        pipeline.get_pipeline_subset_def(solid_selection), run_config=run_config
    )


def define_test_all_scalars_pipeline():
    @lambda_solid(input_defs=[InputDefinition("num", Int)])
    def take_int(num):
        return num

    @lambda_solid(output_def=OutputDefinition(Int))
    def produce_int():
        return 2

    @lambda_solid(input_defs=[InputDefinition("string", String)])
    def take_string(string):
        return string

    @lambda_solid(output_def=OutputDefinition(String))
    def produce_string():
        return "foo"

    @lambda_solid(input_defs=[InputDefinition("float_number", Float)])
    def take_float(float_number):
        return float_number

    @lambda_solid(output_def=OutputDefinition(Float))
    def produce_float():
        return 3.14

    @lambda_solid(input_defs=[InputDefinition("bool_value", Bool)])
    def take_bool(bool_value):
        return bool_value

    @lambda_solid(output_def=OutputDefinition(Bool))
    def produce_bool():
        return True

    @lambda_solid(input_defs=[InputDefinition("any_value", Any)])
    def take_any(any_value):
        return any_value

    @lambda_solid(output_def=OutputDefinition(Any))
    def produce_any():
        return True

    @lambda_solid(input_defs=[InputDefinition("string_list", List[String])])
    def take_string_list(string_list):
        return string_list

    @lambda_solid(input_defs=[InputDefinition("nullable_string", Optional[String])])
    def take_nullable_string(nullable_string):
        return nullable_string

    return PipelineDefinition(
        name="test_all_scalars_pipeline",
        solid_defs=[
            produce_any,
            produce_bool,
            produce_float,
            produce_int,
            produce_string,
            take_any,
            take_bool,
            take_float,
            take_int,
            take_nullable_string,
            take_string,
            take_string_list,
        ],
    )


def single_input_env(solid_name, input_name, input_spec):
    return {"solids": {solid_name: {"inputs": {input_name: input_spec}}}}


def test_int_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        run_config={"solids": {"take_int": {"inputs": {"num": {"value": 2}}}}},
        solid_selection={"take_int"},
    )

    assert result.success
    assert result.result_for_solid("take_int").output_value() == 2


def test_int_input_schema_raw_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        run_config={"solids": {"take_int": {"inputs": {"num": 2}}}},
        solid_selection={"take_int"},
    )

    assert result.success
    assert result.result_for_solid("take_int").output_value() == 2


def test_int_input_schema_failure_wrong_value_type():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_int", "num", {"value": "dkjdfkdj"}),
            solid_selection={"take_int"},
        )
    assert "Error 1: Invalid scalar at path root:solids:take_int:inputs:num:value" in str(
        exc_info.value
    )


def test_int_input_schema_failure_wrong_key():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_int", "num", {"wrong_key": "dkjdfkdj"}),
            solid_selection={"take_int"},
        )
    assert 'Error 1: Undefined field "wrong_key" at path root:solids:take_int:inputs:num' in str(
        exc_info.value
    )


def test_int_input_schema_failure_raw_string():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_int", "num", "dkjdfkdj"),
            solid_selection={"take_int"},
        )
    assert "Error 1: Invalid scalar at path root:solids:take_int:inputs:num" in str(exc_info.value)


def single_output_env(solid_name, output_spec):
    return {"solids": {solid_name: {"outputs": [{"result": output_spec}]}}}


def test_int_json_schema_roundtrip():
    with get_temp_file_name() as tmp_file:
        mat_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_output_env("produce_int", {"json": {"path": tmp_file}}),
            solid_selection={"produce_int"},
        )

        assert mat_result.success

        source_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_int", "num", {"json": {"path": tmp_file}}),
            solid_selection={"take_int"},
        )

        assert source_result.result_for_solid("take_int").output_value() == 2


def test_int_pickle_schema_roundtrip():
    with get_temp_file_name() as tmp_file:
        mat_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_output_env("produce_int", {"pickle": {"path": tmp_file}}),
            solid_selection={"produce_int"},
        )

        assert mat_result.success

        source_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_int", "num", {"pickle": {"path": tmp_file}}),
            solid_selection={"take_int"},
        )

        assert source_result.result_for_solid("take_int").output_value() == 2


def test_string_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        run_config=single_input_env("take_string", "string", {"value": "dkjkfd"}),
        solid_selection={"take_string"},
    )

    assert result.success
    assert result.result_for_solid("take_string").output_value() == "dkjkfd"


def test_string_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_string", "string", {"value": 3343}),
            solid_selection={"take_string"},
        )

    assert "Invalid scalar at path root:solids:take_string:inputs:string:value" in str(
        exc_info.value
    )


def test_float_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        run_config=single_input_env("take_float", "float_number", {"value": 3.34}),
        solid_selection={"take_float"},
    )

    assert result.success
    assert result.result_for_solid("take_float").output_value() == 3.34


def test_float_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_float", "float_number", {"value": "3343"}),
            solid_selection={"take_float"},
        )

    assert "Invalid scalar at path root:solids:take_float:inputs:float_number:value" in str(
        exc_info.value
    )


def test_bool_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        run_config=single_input_env("take_bool", "bool_value", {"value": True}),
        solid_selection={"take_bool"},
    )

    assert result.success
    assert result.result_for_solid("take_bool").output_value() is True


def test_bool_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_bool", "bool_value", {"value": "3343"}),
            solid_selection={"take_bool"},
        )

    assert "Invalid scalar at path root:solids:take_bool:inputs:bool_value:value" in str(
        exc_info.value
    )


def test_any_input_schema_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        run_config=single_input_env("take_any", "any_value", {"value": "ff"}),
        solid_selection={"take_any"},
    )

    assert result.success
    assert result.result_for_solid("take_any").output_value() == "ff"

    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        run_config=single_input_env("take_any", "any_value", {"value": 3843}),
        solid_selection={"take_any"},
    )

    assert result.success
    assert result.result_for_solid("take_any").output_value() == 3843


def test_none_string_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_string", "string", None),
            solid_selection={"take_string"},
        )

    assert len(exc_info.value.errors) == 1

    error = exc_info.value.errors[0]

    assert "Value at path root:solids:take_string:inputs:string must not be None." in error.message


def test_value_none_string_input_schema_failure():
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_string", "string", {"value": None}),
            solid_selection={"take_string"},
        )

    assert "Value at path root:solids:take_string:inputs:string:value must not be None" in str(
        exc_info.value
    )


def test_string_json_schema_roundtrip():
    with get_temp_file_name() as tmp_file:
        mat_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_output_env("produce_string", {"json": {"path": tmp_file}}),
            solid_selection={"produce_string"},
        )

        assert mat_result.success

        source_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_string", "string", {"json": {"path": tmp_file}}),
            solid_selection={"take_string"},
        )

        assert source_result.result_for_solid("take_string").output_value() == "foo"


def test_string_pickle_schema_roundtrip():
    with get_temp_file_name() as tmp_file:
        mat_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_output_env("produce_string", {"pickle": {"path": tmp_file}}),
            solid_selection={"produce_string"},
        )

        assert mat_result.success

        source_result = _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_string", "string", {"pickle": {"path": tmp_file}}),
            solid_selection={"take_string"},
        )

        assert source_result.result_for_solid("take_string").output_value() == "foo"


def test_string_list_input():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        run_config=single_input_env("take_string_list", "string_list", [{"value": "foobar"}]),
        solid_selection={"take_string_list"},
    )

    assert result.success

    assert result.result_for_solid("take_string_list").output_value() == ["foobar"]


def test_nullable_string_input_with_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        run_config=single_input_env("take_nullable_string", "nullable_string", {"value": "foobar"}),
        solid_selection={"take_nullable_string"},
    )

    assert result.success

    assert result.result_for_solid("take_nullable_string").output_value() == "foobar"


def test_nullable_string_input_with_none_value():
    # Perhaps a confusing test case. Optional makes the entire enclosing structure nullable,
    # rather than the "value" value embedded within it
    with pytest.raises(DagsterInvalidConfigError) as exc_info:
        _execute_pipeline_with_subset(
            define_test_all_scalars_pipeline(),
            run_config=single_input_env("take_nullable_string", "nullable_string", {"value": None}),
            solid_selection={"take_nullable_string"},
        )

    assert (
        "Value at path root:solids:take_nullable_string:inputs:nullable_string:value must not be None"
    ) in str(exc_info.value)


def test_nullable_string_input_without_value():
    result = _execute_pipeline_with_subset(
        define_test_all_scalars_pipeline(),
        run_config=single_input_env("take_nullable_string", "nullable_string", None),
        solid_selection={"take_nullable_string"},
    )

    assert result.success

    assert result.result_for_solid("take_nullable_string").output_value() is None



"""Workhorse functions for individual API requests."""

import os
import sys

from dagster import check
from dagster.core.definitions import ScheduleExecutionContext, TriggeredExecutionContext
from dagster.core.definitions.reconstructable import (
    ReconstructablePipeline,
    ReconstructableRepository,
)
from dagster.core.errors import (
    DagsterInvalidSubsetError,
    DagsterSubprocessError,
    PartitionExecutionError,
    ScheduleExecutionError,
    TriggeredExecutionError,
    user_code_error_boundary,
)
from dagster.core.events import EngineEventData
from dagster.core.execution.api import create_execution_plan, execute_run_iterator
from dagster.core.host_representation import external_pipeline_data_from_def
from dagster.core.host_representation.external_data import (
    ExternalExecutionParamsData,
    ExternalExecutionParamsErrorData,
    ExternalPartitionConfigData,
    ExternalPartitionExecutionErrorData,
    ExternalPartitionExecutionParamData,
    ExternalPartitionNamesData,
    ExternalPartitionSetExecutionParamData,
    ExternalPartitionTagsData,
    ExternalPipelineSubsetResult,
    ExternalScheduleExecutionData,
    ExternalScheduleExecutionErrorData,
)
from dagster.core.instance import DagsterInstance
from dagster.core.snap.execution_plan_snapshot import (
    ExecutionPlanSnapshotErrorData,
    snapshot_from_execution_plan,
)
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.grpc.types import (
    ExecutionPlanSnapshotArgs,
    ExternalScheduleExecutionArgs,
    ExternalTriggeredExecutionArgs,
    PartitionArgs,
    PartitionNamesArgs,
    ScheduleExecutionDataMode,
)
from dagster.serdes import deserialize_json_to_dagster_namedtuple
from dagster.serdes.ipc import IPCErrorMessage
from dagster.utils import start_termination_thread
from dagster.utils.error import serializable_error_info_from_exc_info
from dagster.utils.hosted_user_process import recon_repository_from_origin

from .types import ExecuteRunArgs, ExternalScheduleExecutionArgs, PartitionSetExecutionParamArgs


class RunInSubprocessComplete:
    """Sentinel passed over multiprocessing Queue when subprocess is complete"""


class StartRunInSubprocessSuccessful:
    """Sentinel passed over multiprocessing Queue when launch is successful in subprocess."""


def _core_execute_run(recon_pipeline, pipeline_run, instance):
    check.inst_param(recon_pipeline, "recon_pipeline", ReconstructablePipeline)
    check.inst_param(pipeline_run, "pipeline_run", PipelineRun)
    check.inst_param(instance, "instance", DagsterInstance)

    try:
        for event in execute_run_iterator(recon_pipeline, pipeline_run, instance):
            yield event
    except DagsterSubprocessError as err:
        if not all(
            [err_info.cls_name == "KeyboardInterrupt" for err_info in err.subprocess_error_infos]
        ):
            yield instance.report_engine_event(
                "An exception was thrown during execution that is likely a framework error, "
                "rather than an error in user code.",
                pipeline_run,
                EngineEventData.engine_error(serializable_error_info_from_exc_info(sys.exc_info())),
            )
            instance.report_run_failed(pipeline_run)
    except Exception:  # pylint: disable=broad-except
        yield instance.report_engine_event(
            "An exception was thrown during execution that is likely a framework error, "
            "rather than an error in user code.",
            pipeline_run,
            EngineEventData.engine_error(serializable_error_info_from_exc_info(sys.exc_info())),
        )
        instance.report_run_failed(pipeline_run)


def _run_in_subprocess(
    serialized_execute_run_args,
    recon_pipeline,
    termination_event,
    subprocess_status_handler,
    run_event_handler,
):

    start_termination_thread(termination_event)
    try:
        execute_run_args = deserialize_json_to_dagster_namedtuple(serialized_execute_run_args)
        check.inst_param(execute_run_args, "execute_run_args", ExecuteRunArgs)

        instance = DagsterInstance.from_ref(execute_run_args.instance_ref)
        pipeline_run = instance.get_run_by_id(execute_run_args.pipeline_run_id)

        pid = os.getpid()

    except:  # pylint: disable=bare-except
        event = IPCErrorMessage(
            serializable_error_info=serializable_error_info_from_exc_info(sys.exc_info()),
            message="Error during RPC setup for ExecuteRun",
        )
        subprocess_status_handler(event)
        subprocess_status_handler(RunInSubprocessComplete())
        if instance:
            instance.dispose()
        return

    subprocess_status_handler(StartRunInSubprocessSuccessful())

    run_event_handler(
        instance.report_engine_event(
            "Started process for pipeline (pid: {pid}).".format(pid=pid),
            pipeline_run,
            EngineEventData.in_process(pid, marker_end="cli_api_subprocess_init"),
        )
    )

    # This is so nasty but seemingly unavoidable
    # https://amir.rachum.com/blog/2017/03/03/generator-cleanup/
    closed = False
    try:
        for event in _core_execute_run(recon_pipeline, pipeline_run, instance):
            run_event_handler(event)
    except KeyboardInterrupt:
        run_event_handler(
            instance.report_engine_event(
                message="Pipeline execution terminated by interrupt", pipeline_run=pipeline_run,
            )
        )
        raise
    except GeneratorExit:
        closed = True
        raise
    finally:
        if not closed:
            run_event_handler(
                instance.report_engine_event(
                    "Process for pipeline exited (pid: {pid}).".format(pid=pid), pipeline_run,
                )
            )
        subprocess_status_handler(RunInSubprocessComplete())
        instance.dispose()


def execute_run_in_subprocess(
    serialized_execute_run_args, recon_pipeline, event_queue, termination_event
):
    _run_in_subprocess(
        serialized_execute_run_args,
        recon_pipeline,
        termination_event,
        subprocess_status_handler=event_queue.put,
        run_event_handler=event_queue.put,
    )


def start_run_in_subprocess(
    serialized_execute_run_args, recon_pipeline, event_queue, termination_event
):
    _run_in_subprocess(
        serialized_execute_run_args,
        recon_pipeline,
        termination_event,
        subprocess_status_handler=event_queue.put,
        run_event_handler=lambda x: None,
    )


def get_external_pipeline_subset_result(recon_pipeline, solid_selection):
    check.inst_param(recon_pipeline, "recon_pipeline", ReconstructablePipeline)

    if solid_selection:
        try:
            sub_pipeline = recon_pipeline.subset_for_execution(solid_selection)
            definition = sub_pipeline.get_definition()
        except DagsterInvalidSubsetError:
            return ExternalPipelineSubsetResult(
                success=False, error=serializable_error_info_from_exc_info(sys.exc_info())
            )
    else:
        definition = recon_pipeline.get_definition()

    external_pipeline_data = external_pipeline_data_from_def(definition)
    return ExternalPipelineSubsetResult(success=True, external_pipeline_data=external_pipeline_data)


def get_external_schedule_execution(recon_repo, external_schedule_execution_args):
    check.inst_param(
        recon_repo, "recon_repo", ReconstructableRepository,
    )
    check.inst_param(
        external_schedule_execution_args,
        "external_schedule_execution_args",
        ExternalScheduleExecutionArgs,
    )

    definition = recon_repo.get_definition()
    schedule_def = definition.get_schedule_def(external_schedule_execution_args.schedule_name)
    with DagsterInstance.from_ref(external_schedule_execution_args.instance_ref) as instance:
        schedule_context = ScheduleExecutionContext(instance)
        schedule_execution_data_mode = external_schedule_execution_args.schedule_execution_data_mode

        try:
            with user_code_error_boundary(
                ScheduleExecutionError,
                lambda: "Error occurred during the execution of should_execute for schedule "
                "{schedule_name}".format(schedule_name=schedule_def.name),
            ):
                should_execute = None
                if (
                    schedule_execution_data_mode
                    == ScheduleExecutionDataMode.LAUNCH_SCHEDULED_EXECUTION
                ):
                    should_execute = schedule_def.should_execute(schedule_context)
                    if not should_execute:
                        return ExternalScheduleExecutionData(
                            should_execute=False, run_config=None, tags=None
                        )

            with user_code_error_boundary(
                ScheduleExecutionError,
                lambda: "Error occurred during the execution of run_config_fn for schedule "
                "{schedule_name}".format(schedule_name=schedule_def.name),
            ):
                run_config = schedule_def.get_run_config(schedule_context)

            with user_code_error_boundary(
                ScheduleExecutionError,
                lambda: "Error occurred during the execution of tags_fn for schedule "
                "{schedule_name}".format(schedule_name=schedule_def.name),
            ):
                tags = schedule_def.get_tags(schedule_context)

            return ExternalScheduleExecutionData(
                run_config=run_config, tags=tags, should_execute=should_execute
            )
        except ScheduleExecutionError:
            return ExternalScheduleExecutionErrorData(
                serializable_error_info_from_exc_info(sys.exc_info())
            )


def get_external_triggered_execution_params(recon_repo, external_triggered_execution_args):
    check.inst_param(recon_repo, "recon_repo", ReconstructableRepository)
    check.inst_param(
        external_triggered_execution_args,
        "external_triggered_execution_args",
        ExternalTriggeredExecutionArgs,
    )
    definition = recon_repo.get_definition()
    triggered_execution_def = definition.get_triggered_execution_def(
        external_triggered_execution_args.trigger_name
    )
    with DagsterInstance.from_ref(external_triggered_execution_args.instance_ref) as instance:
        context = TriggeredExecutionContext(instance)

        try:
            with user_code_error_boundary(
                TriggeredExecutionError,
                lambda: "Error occured during the execution of should_execute_fn for triggered "
                "execution {name}".format(name=triggered_execution_def.name),
            ):
                should_execute = triggered_execution_def.should_execute(context)
                if not should_execute:
                    return ExternalExecutionParamsData(
                        run_config=None, tags=None, should_execute=False
                    )

            with user_code_error_boundary(
                TriggeredExecutionError,
                lambda: "Error occured during the execution of run_config_fn for triggered "
                "execution {name}".format(name=triggered_execution_def.name),
            ):
                run_config = triggered_execution_def.get_run_config(context)

            with user_code_error_boundary(
                TriggeredExecutionError,
                lambda: "Error occured during the execution of tags_fn for triggered "
                "execution {name}".format(name=triggered_execution_def.name),
            ):
                tags = triggered_execution_def.get_tags(context)

            return ExternalExecutionParamsData(
                run_config=run_config, tags=tags, should_execute=should_execute
            )
        except TriggeredExecutionError:
            return ExternalExecutionParamsErrorData(
                serializable_error_info_from_exc_info(sys.exc_info())
            )


def get_partition_config(args):
    check.inst_param(args, "args", PartitionArgs)
    recon_repo = recon_repository_from_origin(args.repository_origin)
    definition = recon_repo.get_definition()
    partition_set_def = definition.get_partition_set_def(args.partition_set_name)
    partition = partition_set_def.get_partition(args.partition_name)
    try:
        with user_code_error_boundary(
            PartitionExecutionError,
            lambda: "Error occurred during the evaluation of the `run_config_for_partition` "
            "function for partition set {partition_set_name}".format(
                partition_set_name=partition_set_def.name
            ),
        ):
            run_config = partition_set_def.run_config_for_partition(partition)
            return ExternalPartitionConfigData(name=partition.name, run_config=run_config)
    except PartitionExecutionError:
        return ExternalPartitionExecutionErrorData(
            serializable_error_info_from_exc_info(sys.exc_info())
        )


def get_partition_names(args):
    check.inst_param(args, "args", PartitionNamesArgs)
    recon_repo = recon_repository_from_origin(args.repository_origin)
    definition = recon_repo.get_definition()
    partition_set_def = definition.get_partition_set_def(args.partition_set_name)
    try:
        with user_code_error_boundary(
            PartitionExecutionError,
            lambda: "Error occurred during the execution of the partition generation function for "
            "partition set {partition_set_name}".format(partition_set_name=partition_set_def.name),
        ):
            return ExternalPartitionNamesData(
                partition_names=partition_set_def.get_partition_names()
            )
    except PartitionExecutionError:
        return ExternalPartitionExecutionErrorData(
            serializable_error_info_from_exc_info(sys.exc_info())
        )


def get_partition_tags(args):
    check.inst_param(args, "args", PartitionArgs)
    recon_repo = recon_repository_from_origin(args.repository_origin)
    definition = recon_repo.get_definition()
    partition_set_def = definition.get_partition_set_def(args.partition_set_name)
    partition = partition_set_def.get_partition(args.partition_name)
    try:
        with user_code_error_boundary(
            PartitionExecutionError,
            lambda: "Error occurred during the evaluation of the `tags_for_partition` function for "
            "partition set {partition_set_name}".format(partition_set_name=partition_set_def.name),
        ):
            tags = partition_set_def.tags_for_partition(partition)
            return ExternalPartitionTagsData(name=partition.name, tags=tags)
    except PartitionExecutionError:
        return ExternalPartitionExecutionErrorData(
            serializable_error_info_from_exc_info(sys.exc_info())
        )


def get_external_execution_plan_snapshot(recon_pipeline, args):
    check.inst_param(recon_pipeline, "recon_pipeline", ReconstructablePipeline)
    check.inst_param(args, "args", ExecutionPlanSnapshotArgs)

    try:
        pipeline = (
            recon_pipeline.subset_for_execution(args.solid_selection)
            if args.solid_selection
            else recon_pipeline
        )

        return snapshot_from_execution_plan(
            create_execution_plan(
                pipeline=pipeline,
                run_config=args.run_config,
                mode=args.mode,
                step_keys_to_execute=args.step_keys_to_execute,
            ),
            args.pipeline_snapshot_id,
        )
    except:  # pylint: disable=bare-except
        return ExecutionPlanSnapshotErrorData(
            error=serializable_error_info_from_exc_info(sys.exc_info())
        )


def get_partition_set_execution_param_data(args):
    check.inst_param(args, "args", PartitionSetExecutionParamArgs)
    recon_repo = recon_repository_from_origin(args.repository_origin)
    repo_definition = recon_repo.get_definition()
    partition_set_def = repo_definition.get_partition_set_def(args.partition_set_name)
    try:
        with user_code_error_boundary(
            PartitionExecutionError,
            lambda: "Error occurred during the partition generation for partition set "
            "{partition_set_name}".format(partition_set_name=partition_set_def.name),
        ):
            all_partitions = partition_set_def.get_partitions()
        partitions = [
            partition for partition in all_partitions if partition.name in args.partition_names
        ]

        partition_data = []
        for partition in partitions:

            def _error_message_fn(partition_set_name, partition_name):
                return lambda: (
                    "Error occurred during the partition config and tag generation for "
                    "partition set {partition_set_name}::{partition_name}".format(
                        partition_set_name=partition_set_name, partition_name=partition_name
                    )
                )

            with user_code_error_boundary(
                PartitionExecutionError, _error_message_fn(partition_set_def.name, partition.name)
            ):
                run_config = partition_set_def.run_config_for_partition(partition)
                tags = partition_set_def.tags_for_partition(partition)

            partition_data.append(
                ExternalPartitionExecutionParamData(
                    name=partition.name, tags=tags, run_config=run_config,
                )
            )

        return ExternalPartitionSetExecutionParamData(partition_data=partition_data)

    except PartitionExecutionError:
        return ExternalPartitionExecutionErrorData(
            serializable_error_info_from_exc_info(sys.exc_info())
        )



import yaml

from dagster.cli.workspace.config_schema import process_workspace_config
from dagster.core.test_utils import environ


def _validate_yaml_contents(yaml_contents):
    return process_workspace_config(yaml.safe_load(yaml_contents))


def test_repository_yaml_parsing():
    valid_yaml_contents = """
repository:
    module: some_module
    fn: a_repo
    """
    assert _validate_yaml_contents(valid_yaml_contents).success

    invalid_yaml_contents = """
repository:
    module: some_module
    wrong: a_repo
    """

    assert not _validate_yaml_contents(invalid_yaml_contents).success


def test_python_file():
    terse_workspace_yaml = """
load_from:
    - python_file: a_file.py
"""

    assert _validate_yaml_contents(terse_workspace_yaml).success

    nested_workspace_yaml = """
load_from:
    - python_file:
        relative_path: a_file.py
"""

    assert _validate_yaml_contents(nested_workspace_yaml).success

    nested_workspace_yaml_with_def_name = """
load_from:
    - python_file:
        relative_path: a_file.py
        attribute: repo_symbol
"""
    assert _validate_yaml_contents(nested_workspace_yaml_with_def_name).success

    nested_workspace_yaml_with_def_name_and_location = """
load_from:
    - python_file:
        relative_path: a_file.py
        attribute: repo_symbol
        location_name: some_location
"""
    assert _validate_yaml_contents(nested_workspace_yaml_with_def_name_and_location).success


def test_python_module():
    terse_workspace_yaml = """
load_from:
    - python_module: a_module
"""

    assert _validate_yaml_contents(terse_workspace_yaml).success

    nested_workspace_yaml = """
load_from:
    - python_module:
        module_name: a_module
"""

    assert _validate_yaml_contents(nested_workspace_yaml).success

    nested_workspace_yaml_with_def_name = """
load_from:
    - python_module:
        module_name: a_module
        attribute: repo_symbol
"""
    assert _validate_yaml_contents(nested_workspace_yaml_with_def_name).success

    nested_workspace_yaml_with_def_name_and_location = """
load_from:
    - python_module:
        module_name: a_module
        attribute: repo_symbol
        location_name: some_location
"""
    assert _validate_yaml_contents(nested_workspace_yaml_with_def_name_and_location).success


def test_cannot_do_both():
    both_yaml = """
load_from:
    - python_module: a_module
      python_file: a_file.py
"""
    assert not _validate_yaml_contents(both_yaml).success


def test_load_both():
    both_yaml = """
load_from:
    - python_module: a_module
    - python_file: a_file.py
"""

    assert _validate_yaml_contents(both_yaml).success


def test_load_python_environment_with_file():
    python_environment_yaml_with_file = """
load_from:
    - python_environment:
        executable_path: /path/to/venv/bin/python
        target:
            python_file: file_valid_in_that_env.py
"""

    validation_result = _validate_yaml_contents(python_environment_yaml_with_file)

    assert validation_result.success


def test_load_python_environment_with_module():
    python_environment_yaml_with_module = """
load_from:
    - python_environment:
        executable_path: /path/to/venv/bin/python
        target:
            python_module: module_valid_in_that_env.py
"""

    validation_result = _validate_yaml_contents(python_environment_yaml_with_module)

    assert validation_result.success


def test_load_python_environment_with_env_var():
    with environ({"TEST_EXECUTABLE_PATH": "executable/path/bin/python"}):
        python_environment_yaml_with_file = """
    load_from:
        - python_environment:
            executable_path:
                env: TEST_EXECUTABLE_PATH
            target:
                python_file: file_valid_in_that_env.py
    """

        validation_result = _validate_yaml_contents(python_environment_yaml_with_file)

        assert validation_result.success


def test_load_from_grpc_server():
    with environ({"TEST_EXECUTABLE_PATH": "executable/path/bin/python"}):
        valid_yaml = """
    load_from:
        - grpc_server:
            host: remotehost
            port: 4266
            location_name: 'my_grpc_server'
    """

        validation_result = _validate_yaml_contents(valid_yaml)

        assert validation_result.success


def test_load_python_environment_and_grpc_server():
    with environ({"TEST_EXECUTABLE_PATH": "executable/path/bin/python"}):
        valid_yaml = """
    load_from:
        - grpc_server:
            host: remotehost
            port: 4266
            location_name: 'my_grpc_server'
        - python_environment:
            executable_path:
                env: TEST_EXECUTABLE_PATH
            target:
                python_file: file_valid_in_that_env.py
    """

        validation_result = _validate_yaml_contents(valid_yaml)

        assert validation_result.success



import datetime

from dateutil.relativedelta import relativedelta

from dagster import check
from dagster.core.errors import DagsterInvariantViolationError

DEFAULT_DATE_FORMAT = "%Y-%m-%d"


def date_partition_range(
    start, end=None, delta=datetime.timedelta(days=1), fmt=None, inclusive=False
):
    """ Utility function that returns a partition generating function to be used in creating a
    `PartitionSet` definition.

    Args:
        start (datetime): Datetime capturing the start of the time range.
        end  (Optional(datetime)): Datetime capturing the end of the partition.  By default, the
                                   current time is used.  The range is not inclusive of the end
                                   value.
        delta (Optional(timedelta)): Timedelta representing the time duration of each partition.
        fmt (Optional(str)): Format string to represent each partition by its start time
        inclusive (Optional(bool)): By default, the partition set only contains date interval
            partitions for which the end time of the interval is less than current time. In other
            words, the partition set contains date interval partitions that are completely in the
            past. If inclusive is set to True, then the partition set will include all date
            interval partitions for which the start time of the interval is less than the
            current time.

    Returns:
        Callable[[], List[Partition]]
    """
    from dagster.core.definitions.partition import Partition

    check.inst_param(start, "start", datetime.datetime)
    check.opt_inst_param(end, "end", datetime.datetime)
    check.inst_param(delta, "timedelta", (datetime.timedelta, relativedelta))
    fmt = check.opt_str_param(fmt, "fmt", default=DEFAULT_DATE_FORMAT)

    if end and start > end:
        raise DagsterInvariantViolationError(
            'Selected date range start "{start}" is after date range end "{end}'.format(
                start=start.strftime(fmt), end=end.strftime(fmt)
            )
        )

    def get_date_range_partitions():
        current = start

        _end_specified = end is not None
        _end = end or datetime.datetime.now()

        date_names = []
        while current < _end or (_end_specified and current == _end):
            date_names.append(Partition(value=current, name=current.strftime(fmt)))
            current = current + delta

        # We don't include the last element here by default since we only want
        # fully completed intervals, and the _end time is in the middle of the interval
        # represented by the last element of date_names
        if inclusive:
            return date_names

        return date_names[:-1]

    return get_date_range_partitions



from dagster import Field, Int, Noneable, PipelineDefinition, ScalarUnion, String, solid
from dagster.config.field import resolve_to_config_type
from dagster.config.iterate_types import config_schema_snapshot_from_config_type
from dagster.config.snap import get_recursive_type_keys, snap_from_config_type
from dagster.config.type_printer import print_config_type_to_string


def assert_inner_types(parent_type, *dagster_types):
    config_type = resolve_to_config_type(parent_type)
    config_schema_snapshot = config_schema_snapshot_from_config_type(config_type)

    all_type_keys = get_recursive_type_keys(
        snap_from_config_type(config_type), config_schema_snapshot
    )

    assert set(all_type_keys) == set(
        map(lambda x: x.key, map(resolve_to_config_type, dagster_types))
    )


def test_basic_type_print():
    assert print_config_type_to_string(Int) == "Int"
    assert_inner_types(Int)


def test_basic_list_type_print():
    assert print_config_type_to_string([int]) == "[Int]"
    assert_inner_types([int], Int)


def test_double_list_type_print():
    assert print_config_type_to_string([[int]]) == "[[Int]]"
    int_list = [int]
    list_int_list = [int_list]
    assert_inner_types(list_int_list, Int, int_list)


def test_basic_nullable_type_print():
    assert print_config_type_to_string(Noneable(int)) == "Int?"
    nullable_int = Noneable(int)
    assert_inner_types(nullable_int, Int)


def test_nullable_list_combos():
    assert print_config_type_to_string([int]) == "[Int]"
    assert print_config_type_to_string(Noneable([int])) == "[Int]?"
    assert print_config_type_to_string([Noneable(int)]) == "[Int?]"
    assert print_config_type_to_string(Noneable([Noneable(int)])) == "[Int?]?"


def test_basic_dict():
    output = print_config_type_to_string({"int_field": int})

    expected = """{
  int_field: Int
}"""

    assert output == expected


def test_two_field_dicts():
    two_field_dict = {"int_field": int, "string_field": str}
    assert_inner_types(two_field_dict, Int, String)

    output = print_config_type_to_string(two_field_dict)

    expected = """{
  int_field: Int
  string_field: String
}"""

    assert output == expected


def test_two_field_dicts_same_type():
    two_field_dict = {"int_field1": int, "int_field2": int}
    assert_inner_types(two_field_dict, Int)

    output = print_config_type_to_string(two_field_dict)

    expected = """{
  int_field1: Int
  int_field2: Int
}"""

    assert output == expected


def test_optional_field():
    output = print_config_type_to_string({"int_field": Field(int, is_required=False)})

    expected = """{
  int_field?: Int
}"""

    assert output == expected


def test_single_level_dict_lists_and_nullable():
    output = print_config_type_to_string(
        {
            "nullable_int_field": Noneable(int),
            "optional_int_field": Field(int, is_required=False),
            "string_list_field": [str],
        }
    )

    expected = """{
  nullable_int_field: Int?
  optional_int_field?: Int
  string_list_field: [String]
}"""

    assert output == expected


def test_nested_dict():
    nested_type = {"int_field": int}
    outer_type = {"nested": nested_type}
    output = print_config_type_to_string(outer_type)

    assert_inner_types(outer_type, Int, nested_type)

    expected = """{
  nested: {
    int_field: Int
  }
}"""

    assert output == expected


def test_scalar_union():
    non_scalar_type = {"str_field": String}
    scalar_union_type = ScalarUnion(scalar_type=int, non_scalar_schema=non_scalar_type,)
    assert_inner_types(scalar_union_type, String, Int, non_scalar_type)


def test_test_type_pipeline_construction():
    assert define_test_type_pipeline()


def define_solid_for_test_type(name, config):
    @solid(name=name, config_schema=config, input_defs=[], output_defs=[])
    def a_solid(_):
        return None

    return a_solid


# launch in dagit with this command:
# dagit -f test_type_printer.py -n define_test_type_pipeline
def define_test_type_pipeline():
    return PipelineDefinition(
        name="test_type_pipeline",
        solid_defs=[
            define_solid_for_test_type("int_config", int),
            define_solid_for_test_type("list_of_int_config", [int]),
            define_solid_for_test_type("nullable_list_of_int_config", Noneable([int])),
            define_solid_for_test_type("list_of_nullable_int_config", [Noneable(int)]),
            define_solid_for_test_type(
                "nullable_list_of_nullable_int_config", Noneable([Noneable(int)])
            ),
            define_solid_for_test_type("simple_dict", {"int_field": int, "string_field": str}),
            define_solid_for_test_type(
                "dict_with_optional_field",
                {
                    "nullable_int_field": Noneable(int),
                    "optional_int_field": Field(int, is_required=False),
                    "string_list_field": [str],
                },
            ),
            define_solid_for_test_type("nested_dict", {"nested": {"int_field": int}}),
        ],
    )



from dagster_tests.general_tests.test_repository import define_multi_mode_with_resources_pipeline

from dagster import PresetDefinition
from dagster.utils import file_relative_path


def test_preset_yaml_roundtrip():
    pipeline = define_multi_mode_with_resources_pipeline()

    preset_def = pipeline.get_preset("add")

    with open(
        file_relative_path(__file__, "../environments/multi_mode_with_resources/add_mode.yaml"), "r"
    ) as fd:
        assert preset_def.get_environment_yaml() == fd.read()


def test_empty_preset():
    empty_preset = PresetDefinition("empty")
    assert empty_preset.run_config == None
    assert empty_preset.get_environment_yaml() == "{}\n"



import re

import pytest

from dagster import (
    Bool,
    DagsterInstance,
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    ModeDefinition,
    PipelineDefinition,
    PresetDefinition,
    check,
    execute_pipeline,
    lambda_solid,
    pipeline,
    solid,
)
from dagster.utils import file_relative_path


def test_presets():
    @solid(config_schema={"error": Bool})
    def can_fail(context):
        if context.solid_config["error"]:
            raise Exception("I did an error")
        return "cool"

    @lambda_solid
    def always_fail():
        raise Exception("I always do this")

    pipe = PipelineDefinition(
        name="simple",
        solid_defs=[can_fail, always_fail],
        preset_defs=[
            PresetDefinition.from_files(
                "passing",
                config_files=[file_relative_path(__file__, "pass_env.yaml")],
                solid_selection=["can_fail"],
            ),
            PresetDefinition.from_files(
                "passing_overide_to_fail",
                config_files=[file_relative_path(__file__, "pass_env.yaml")],
                solid_selection=["can_fail"],
            ).with_additional_config({"solids": {"can_fail": {"config": {"error": True}}}}),
            PresetDefinition(
                "passing_direct_dict",
                run_config={"solids": {"can_fail": {"config": {"error": False}}}},
                solid_selection=["can_fail"],
            ),
            PresetDefinition.from_files(
                "failing_1",
                config_files=[file_relative_path(__file__, "fail_env.yaml")],
                solid_selection=["can_fail"],
            ),
            PresetDefinition.from_files(
                "failing_2", config_files=[file_relative_path(__file__, "pass_env.yaml")]
            ),
            PresetDefinition("subset", solid_selection=["can_fail"],),
        ],
    )

    with pytest.raises(DagsterInvariantViolationError):
        PresetDefinition.from_files(
            "invalid_1", config_files=[file_relative_path(__file__, "not_a_file.yaml")]
        )

    with pytest.raises(DagsterInvariantViolationError):
        PresetDefinition.from_files(
            "invalid_2",
            config_files=[file_relative_path(__file__, "test_repository_definition.py")],
        )

    assert execute_pipeline(pipe, preset="passing").success

    assert execute_pipeline(pipe, preset="passing_direct_dict").success
    assert execute_pipeline(pipe, preset="failing_1", raise_on_error=False).success == False

    assert execute_pipeline(pipe, preset="failing_2", raise_on_error=False).success == False

    with pytest.raises(DagsterInvariantViolationError, match="Could not find preset"):
        execute_pipeline(pipe, preset="not_failing", raise_on_error=False)

    assert (
        execute_pipeline(pipe, preset="passing_overide_to_fail", raise_on_error=False).success
        == False
    )

    assert execute_pipeline(
        pipe, preset="passing", run_config={"solids": {"can_fail": {"config": {"error": False}}}},
    ).success

    with pytest.raises(
        check.CheckError,
        match=re.escape(
            "The environment set in preset 'passing' does not agree with the environment passed "
            "in the `run_config` argument."
        ),
    ):
        execute_pipeline(
            pipe,
            preset="passing",
            run_config={"solids": {"can_fail": {"config": {"error": True}}}},
        )

    assert execute_pipeline(
        pipe, preset="subset", run_config={"solids": {"can_fail": {"config": {"error": False}}}},
    ).success


def test_invalid_preset():
    @lambda_solid
    def lil_solid():
        return ";)"

    with pytest.raises(DagsterInvalidDefinitionError, match='mode "mode_b" which is not defined'):
        PipelineDefinition(
            name="preset_modes",
            solid_defs=[lil_solid],
            mode_defs=[ModeDefinition(name="mode_a")],
            preset_defs=[PresetDefinition(name="preset_b", mode="mode_b")],
        )


def test_conflicting_preset():
    @lambda_solid
    def lil_solid():
        return ";)"

    with pytest.raises(
        DagsterInvalidDefinitionError, match="Two PresetDefinitions seen with the name"
    ):
        PipelineDefinition(
            name="preset_modes",
            solid_defs=[lil_solid],
            mode_defs=[ModeDefinition(name="mode_a")],
            preset_defs=[
                PresetDefinition(name="preset_a", mode="mode_a"),
                PresetDefinition(name="preset_a", mode="mode_a"),
            ],
        )


def test_from_yaml_strings():
    a = """
foo:
  bar: 1
baz: 3
"""
    b = """
foo:
  one: "one"
other: 4
"""
    c = """
final: "result"
"""
    preset = PresetDefinition.from_yaml_strings("passing", [a, b, c])
    assert preset.run_config == {
        "foo": {"bar": 1, "one": "one"},
        "baz": 3,
        "other": 4,
        "final": "result",
    }
    with pytest.raises(
        DagsterInvariantViolationError, match="Encountered error attempting to parse yaml"
    ):
        PresetDefinition.from_yaml_strings("failing", ["--- `"])

    res = PresetDefinition.from_yaml_strings("empty")
    assert res == PresetDefinition(
        name="empty", run_config={}, solid_selection=None, mode="default"
    )


def test_from_pkg_resources():
    good = ("dagster_tests.core_tests.definitions_tests", "pass_env.yaml")
    res = PresetDefinition.from_pkg_resources("pass", [good])
    assert res.run_config == {"solids": {"can_fail": {"config": {"error": False}}}}

    bad_defs = [
        ("dagster_tests.core_tests.definitions_tests", "does_not_exist.yaml"),
        ("dagster_tests.core_tests.definitions_tests", "bad_file_binary.yaml"),
        ("dagster_tests.core_tests.does_not_exist", "some_file.yaml"),
    ]

    for bad_def in bad_defs:
        with pytest.raises(
            DagsterInvariantViolationError, match="Encountered error attempting to parse yaml",
        ):
            PresetDefinition.from_pkg_resources("bad_def", [bad_def])


def test_tags():
    @lambda_solid
    def a_solid():
        return "solid"

    @pipeline(
        tags={"pipeline_tag": "pipeline_tag", "all": "pipeline", "defs": "pipeline"},
        preset_defs=[
            PresetDefinition(
                name="a_preset",
                tags={"preset_tag": "preset_tag", "all": "preset", "defs": "preset"},
            )
        ],
    )
    def a_pipeline():
        a_solid()

    instance = DagsterInstance.ephemeral()
    result = execute_pipeline(
        a_pipeline,
        instance=instance,
        preset="a_preset",
        tags={"execute_tag": "execute_tag", "all": "execute"},
    )
    assert result.success
    pipeline_run = instance.get_run_by_id(result.run_id)

    assert "pipeline_tag" in pipeline_run.tags
    assert "preset_tag" in pipeline_run.tags
    assert "execute_tag" in pipeline_run.tags

    # execute overwrites preset overwrites pipeline def
    assert "all" in pipeline_run.tags
    assert pipeline_run.tags["all"] == "execute"

    # preset overwrites pipeline def
    assert "defs" in pipeline_run.tags
    assert pipeline_run.tags["defs"] == "preset"



from __future__ import print_function

import json
import os
import string
from contextlib import contextmanager

import mock
import pytest
from click.testing import CliRunner

from dagster import (
    PartitionSetDefinition,
    PresetDefinition,
    ScheduleDefinition,
    Shape,
    check,
    lambda_solid,
    pipeline,
    repository,
    seven,
    solid,
)
from dagster.cli.run import run_list_command, run_wipe_command
from dagster.core.host_representation import ExternalPipeline
from dagster.core.instance import DagsterInstance
from dagster.core.launcher import RunLauncher
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.core.test_utils import instance_for_test_tempdir
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin
from dagster.grpc.server import GrpcServerProcess
from dagster.serdes import ConfigurableClass
from dagster.utils import file_relative_path, merge_dicts


def no_print(_):
    return None


@lambda_solid
def do_something():
    return 1


@lambda_solid
def do_input(x):
    return x


@pipeline(
    name="foo", preset_defs=[PresetDefinition(name="test", tags={"foo": "bar"}),],
)
def foo_pipeline():
    do_input(do_something())


def define_foo_pipeline():
    return foo_pipeline


@pipeline(name="baz", description="Not much tbh")
def baz_pipeline():
    do_input()


def not_a_repo_or_pipeline_fn():
    return "kdjfkjdf"


not_a_repo_or_pipeline = 123


@pipeline
def partitioned_scheduled_pipeline():
    do_something()


def define_bar_schedules():
    partition_set = PartitionSetDefinition(
        name="scheduled_partitions",
        pipeline_name="partitioned_scheduled_pipeline",
        partition_fn=lambda: string.digits,
    )
    return {
        "foo_schedule": ScheduleDefinition(
            "foo_schedule", cron_schedule="* * * * *", pipeline_name="test_pipeline", run_config={},
        ),
        "partitioned_schedule": partition_set.create_schedule_definition(
            schedule_name="partitioned_schedule", cron_schedule="* * * * *"
        ),
    }


def define_bar_partitions():
    def error_name():
        raise Exception("womp womp")

    def error_config(_):
        raise Exception("womp womp")

    return {
        "baz_partitions": PartitionSetDefinition(
            name="baz_partitions",
            pipeline_name="baz",
            partition_fn=lambda: string.digits,
            run_config_fn_for_partition=lambda partition: {
                "solids": {"do_input": {"inputs": {"x": {"value": partition.value}}}}
            },
        ),
        "error_name_partitions": PartitionSetDefinition(
            name="error_name_partitions", pipeline_name="baz", partition_fn=error_name,
        ),
        "error_config_partitions": PartitionSetDefinition(
            name="error_config_partitions", pipeline_name="baz", partition_fn=error_config,
        ),
    }


@repository
def bar():
    return {
        "pipelines": {
            "foo": foo_pipeline,
            "baz": baz_pipeline,
            "partitioned_scheduled_pipeline": partitioned_scheduled_pipeline,
        },
        "schedules": define_bar_schedules(),
        "partition_sets": define_bar_partitions(),
    }


@solid
def spew(context):
    context.log.info("HELLO WORLD")


@solid
def fail(context):
    raise Exception("I AM SUPPOSED TO FAIL")


@pipeline
def stdout_pipeline():
    spew()


@pipeline
def stderr_pipeline():
    fail()


class InMemoryRunLauncher(RunLauncher, ConfigurableClass):
    def __init__(self, inst_data=None):
        self._inst_data = inst_data
        self._queue = []

    def launch_run(self, instance, run, external_pipeline):
        check.inst_param(instance, "instance", DagsterInstance)
        check.inst_param(run, "run", PipelineRun)
        check.inst_param(external_pipeline, "external_pipeline", ExternalPipeline)
        self._queue.append(run)
        return run

    def queue(self):
        return self._queue

    @classmethod
    def config_type(cls):
        return Shape({})

    @classmethod
    def from_config_value(cls, inst_data, config_value):
        return cls(inst_data=inst_data,)

    @property
    def inst_data(self):
        return self._inst_data

    def can_terminate(self, run_id):
        return False

    def terminate(self, run_id):
        check.not_implemented("Termintation not supported")


@contextmanager
def default_instance_tempdir(temp_dir, overrides=None):
    default_overrides = {
        "run_launcher": {
            "module": "dagster_tests.cli_tests.command_tests.test_cli_commands",
            "class": "InMemoryRunLauncher",
        }
    }
    with instance_for_test_tempdir(
        temp_dir, overrides=merge_dicts(default_overrides, (overrides if overrides else {}))
    ) as instance:
        with mock.patch("dagster.core.instance.DagsterInstance.get") as _instance:
            _instance.return_value = instance
            yield instance


@contextmanager
def default_instance(overrides=None):
    with seven.TemporaryDirectory() as temp_dir:
        with default_instance_tempdir(temp_dir, overrides) as instance:
            yield instance


def managed_grpc_instance():
    return default_instance(overrides={"opt_in": {"local_servers": True}})


@contextmanager
def args_with_instance(gen_instance, *args):
    with gen_instance as instance:
        yield args + (instance,)


def args_with_default_instance(*args):
    return args_with_instance(default_instance(), *args)


def args_with_managed_grpc_instance(*args):
    return args_with_instance(managed_grpc_instance(), *args)


@contextmanager
def grpc_server_bar_kwargs(pipeline_name=None):
    server_process = GrpcServerProcess(
        loadable_target_origin=LoadableTargetOrigin(
            python_file=file_relative_path(__file__, "test_cli_commands.py"), attribute="bar"
        ),
    )
    with server_process.create_ephemeral_client() as client:
        args = {"grpc_host": client.host}
        if pipeline_name:
            args["pipeline"] = "foo"
        if client.port:
            args["grpc_port"] = client.port
        if client.socket:
            args["grpc_socket"] = client.socket
        yield args
    server_process.wait()


@contextmanager
def python_bar_cli_args(pipeline_name=None):
    args = [
        "-m",
        "dagster_tests.cli_tests.command_tests.test_cli_commands",
        "-a",
        "bar",
    ]
    if pipeline_name:
        args.append("-p")
        args.append(pipeline_name)
    yield args


@contextmanager
def grpc_server_bar_cli_args(pipeline_name=None):
    server_process = GrpcServerProcess(
        loadable_target_origin=LoadableTargetOrigin(
            python_file=file_relative_path(__file__, "test_cli_commands.py"), attribute="bar"
        ),
    )
    with server_process.create_ephemeral_client() as client:
        args = ["--grpc-host", client.host]
        if client.port:
            args.append("--grpc-port")
            args.append(client.port)
        if client.socket:
            args.append("--grpc-socket")
            args.append(client.socket)
        if pipeline_name:
            args.append("--pipeline")
            args.append(pipeline_name)

        yield args
    server_process.wait()


@contextmanager
def grpc_server_bar_pipeline_args():
    with default_instance() as instance:
        with grpc_server_bar_kwargs(pipeline_name="foo") as kwargs:
            yield kwargs, False, instance


# This iterates over a list of contextmanagers that can be used to contruct
# (cli_args, uses_legacy_repository_yaml_format, instance tuples)
def launch_command_contexts():
    for pipeline_target_args in valid_external_pipeline_target_args():
        yield args_with_default_instance(*pipeline_target_args)
        yield pytest.param(
            args_with_managed_grpc_instance(*pipeline_target_args), marks=pytest.mark.managed_grpc
        )
    yield pytest.param(grpc_server_bar_pipeline_args(), marks=pytest.mark.deployed_grpc)


def pipeline_python_origin_contexts():
    return [
        args_with_default_instance(pipeline_target_args)
        for pipeline_target_args in valid_pipeline_python_origin_target_args()
    ]


@contextmanager
def scheduler_instance(overrides=None):
    with seven.TemporaryDirectory() as temp_dir:
        with default_instance_tempdir(
            temp_dir,
            overrides=merge_dicts(
                {
                    "scheduler": {
                        "module": "dagster.utils.test",
                        "class": "FilesystemTestScheduler",
                        "config": {"base_dir": temp_dir},
                    }
                },
                overrides if overrides else {},
            ),
        ) as instance:
            yield instance


def managed_grpc_scheduler_instance():
    return scheduler_instance(overrides={"opt_in": {"local_servers": True}})


@contextmanager
def grpc_server_scheduler_cli_args():
    with scheduler_instance() as instance:
        with grpc_server_bar_cli_args() as args:
            yield args, instance


# Returns a list of contextmanagers that can be used to contruct
# (cli_args, instance) tuples for schedule calls
def schedule_command_contexts():
    return [
        args_with_instance(
            scheduler_instance(), ["-w", file_relative_path(__file__, "workspace.yaml")]
        ),
        pytest.param(
            args_with_instance(
                managed_grpc_scheduler_instance(),
                ["-w", file_relative_path(__file__, "workspace.yaml")],
            ),
            marks=pytest.mark.managed_grpc,
        ),
        pytest.param(grpc_server_scheduler_cli_args(), marks=pytest.mark.deployed_grpc),
    ]


# This iterates over a list of contextmanagers that can be used to contruct
# (cli_args, uses_legacy_repository_yaml_format, instance) tuples for backfill calls
def backfill_command_contexts():
    repo_args = {
        "noprompt": True,
        "workspace": (file_relative_path(__file__, "repository_file.yaml"),),
    }
    return [
        args_with_instance(default_instance(), repo_args, True),
        pytest.param(
            args_with_instance(managed_grpc_instance(), repo_args, True),
            marks=pytest.mark.managed_grpc,
        ),
        pytest.param(grpc_server_backfill_args(), marks=pytest.mark.deployed_grpc),
    ]


@contextmanager
def grpc_server_backfill_args():
    with default_instance() as instance:
        with grpc_server_bar_kwargs() as args:
            yield merge_dicts(args, {"noprompt": True}), False, instance


def valid_pipeline_python_origin_target_args():
    return [
        {
            "workspace": None,
            "pipeline": "foo",
            "python_file": file_relative_path(__file__, "test_cli_commands.py"),
            "module_name": None,
            "attribute": "bar",
        },
        {
            "workspace": None,
            "pipeline": "foo",
            "python_file": file_relative_path(__file__, "test_cli_commands.py"),
            "module_name": None,
            "attribute": "bar",
            "working_directory": os.path.dirname(__file__),
        },
        {
            "workspace": None,
            "pipeline": "foo",
            "python_file": None,
            "module_name": "dagster_tests.cli_tests.command_tests.test_cli_commands",
            "attribute": "bar",
        },
        {
            "workspace": None,
            "pipeline": None,
            "python_file": None,
            "module_name": "dagster_tests.cli_tests.command_tests.test_cli_commands",
            "attribute": "foo_pipeline",
        },
        {
            "workspace": None,
            "pipeline": None,
            "python_file": file_relative_path(__file__, "test_cli_commands.py"),
            "module_name": None,
            "attribute": "define_foo_pipeline",
        },
        {
            "workspace": None,
            "pipeline": None,
            "python_file": file_relative_path(__file__, "test_cli_commands.py"),
            "module_name": None,
            "attribute": "define_foo_pipeline",
            "working_directory": os.path.dirname(__file__),
        },
        {
            "workspace": None,
            "pipeline": None,
            "python_file": file_relative_path(__file__, "test_cli_commands.py"),
            "module_name": None,
            "attribute": "foo_pipeline",
        },
    ]


# [(cli_args, uses_legacy_repository_yaml_format)]
def valid_external_pipeline_target_args():
    return [
        (
            {
                "workspace": (file_relative_path(__file__, "repository_file.yaml"),),
                "pipeline": "foo",
                "python_file": None,
                "module_name": None,
                "attribute": None,
            },
            True,
        ),
        (
            {
                "workspace": (file_relative_path(__file__, "repository_module.yaml"),),
                "pipeline": "foo",
                "python_file": None,
                "module_name": None,
                "attribute": None,
            },
            True,
        ),
    ] + [(args, False) for args in valid_pipeline_python_origin_target_args()]


def valid_pipeline_python_origin_target_cli_args():
    return [
        ["-f", file_relative_path(__file__, "test_cli_commands.py"), "-a", "bar", "-p", "foo"],
        [
            "-f",
            file_relative_path(__file__, "test_cli_commands.py"),
            "-d",
            os.path.dirname(__file__),
            "-a",
            "bar",
            "-p",
            "foo",
        ],
        [
            "-m",
            "dagster_tests.cli_tests.command_tests.test_cli_commands",
            "-a",
            "bar",
            "-p",
            "foo",
        ],
        ["-m", "dagster_tests.cli_tests.command_tests.test_cli_commands", "-a", "foo_pipeline"],
        ["-f", file_relative_path(__file__, "test_cli_commands.py"), "-a", "define_foo_pipeline",],
        [
            "-f",
            file_relative_path(__file__, "test_cli_commands.py"),
            "-d",
            os.path.dirname(__file__),
            "-a",
            "define_foo_pipeline",
        ],
    ]


# [(cli_args, uses_legacy_repository_yaml_format)]
def valid_external_pipeline_target_cli_args_no_preset():
    return [
        (["-w", file_relative_path(__file__, "repository_file.yaml"), "-p", "foo"], True),
        (["-w", file_relative_path(__file__, "repository_module.yaml"), "-p", "foo"], True),
        (["-w", file_relative_path(__file__, "workspace.yaml"), "-p", "foo"], False),
        (
            [
                "-w",
                file_relative_path(__file__, "override.yaml"),
                "-w",
                file_relative_path(__file__, "workspace.yaml"),
                "-p",
                "foo",
            ],
            False,
        ),
    ] + [(args, False) for args in valid_pipeline_python_origin_target_cli_args()]


def valid_external_pipeline_target_cli_args_with_preset():
    run_config = {"storage": {"filesystem": {"config": {"base_dir": "/tmp"}}}}
    return valid_external_pipeline_target_cli_args_no_preset() + [
        (
            [
                "-f",
                file_relative_path(__file__, "test_cli_commands.py"),
                "-d",
                os.path.dirname(__file__),
                "-a",
                "define_foo_pipeline",
                "--preset",
                "test",
            ],
            False,
        ),
        (
            [
                "-f",
                file_relative_path(__file__, "test_cli_commands.py"),
                "-d",
                os.path.dirname(__file__),
                "-a",
                "define_foo_pipeline",
                "--config-json",
                json.dumps(run_config),
            ],
            False,
        ),
    ]


def test_run_list():
    runner = CliRunner()
    result = runner.invoke(run_list_command)
    assert result.exit_code == 0


def test_run_wipe_correct_delete_message():
    runner = CliRunner()
    result = runner.invoke(run_wipe_command, input="DELETE\n")
    assert "Deleted all run history and event logs" in result.output
    assert result.exit_code == 0


def test_run_wipe_incorrect_delete_message():
    runner = CliRunner()
    result = runner.invoke(run_wipe_command, input="WRONG\n")
    assert "Exiting without deleting all run history and event logs" in result.output
    assert result.exit_code == 0



import os
import types

import pytest
import yaml

from dagster import (
    DagsterEventType,
    DagsterInvalidConfigError,
    InputDefinition,
    OutputDefinition,
    PipelineRun,
    check,
    execute_pipeline,
    pipeline,
    seven,
    solid,
)
from dagster.core.execution.stats import StepEventStatus
from dagster.core.instance import DagsterInstance, InstanceRef, InstanceType
from dagster.core.launcher import CliApiRunLauncher
from dagster.core.storage.event_log import SqliteEventLogStorage
from dagster.core.storage.local_compute_log_manager import LocalComputeLogManager
from dagster.core.storage.pipeline_run import PipelineRunStatus
from dagster.core.storage.root import LocalArtifactStorage
from dagster.core.storage.runs import SqliteRunStorage


def test_fs_stores():
    @pipeline
    def simple():
        @solid
        def easy(context):
            context.log.info("easy")
            return "easy"

        easy()

    with seven.TemporaryDirectory() as temp_dir:
        run_store = SqliteRunStorage.from_local(temp_dir)
        event_store = SqliteEventLogStorage(temp_dir)
        compute_log_manager = LocalComputeLogManager(temp_dir)
        instance = DagsterInstance(
            instance_type=InstanceType.PERSISTENT,
            local_artifact_storage=LocalArtifactStorage(temp_dir),
            run_storage=run_store,
            event_storage=event_store,
            compute_log_manager=compute_log_manager,
            run_launcher=CliApiRunLauncher(),
        )

        result = execute_pipeline(simple, instance=instance)

        assert run_store.has_run(result.run_id)
        assert run_store.get_run_by_id(result.run_id).status == PipelineRunStatus.SUCCESS
        assert DagsterEventType.PIPELINE_SUCCESS in [
            event.dagster_event.event_type
            for event in event_store.get_logs_for_run(result.run_id)
            if event.is_dagster_event
        ]
        stats = event_store.get_stats_for_run(result.run_id)
        assert stats.steps_succeeded == 1
        assert stats.end_time is not None


def test_init_compute_log_with_bad_config():
    with seven.TemporaryDirectory() as tmpdir_path:
        with open(os.path.join(tmpdir_path, "dagster.yaml"), "w") as fd:
            yaml.dump({"compute_logs": {"garbage": "flargh"}}, fd, default_flow_style=False)
        with pytest.raises(DagsterInvalidConfigError, match='Undefined field "garbage"'):
            DagsterInstance.from_ref(InstanceRef.from_dir(tmpdir_path))


def test_init_compute_log_with_bad_config_override():
    with seven.TemporaryDirectory() as tmpdir_path:
        with pytest.raises(DagsterInvalidConfigError, match='Undefined field "garbage"'):
            DagsterInstance.from_ref(
                InstanceRef.from_dir(tmpdir_path, overrides={"compute_logs": {"garbage": "flargh"}})
            )


def test_init_compute_log_with_bad_config_module():
    with seven.TemporaryDirectory() as tmpdir_path:
        with open(os.path.join(tmpdir_path, "dagster.yaml"), "w") as fd:
            yaml.dump(
                {"compute_logs": {"module": "flargh", "class": "Woble", "config": {}}},
                fd,
                default_flow_style=False,
            )
        with pytest.raises(check.CheckError, match="Couldn't import module"):
            DagsterInstance.from_ref(InstanceRef.from_dir(tmpdir_path))


MOCK_HAS_RUN_CALLED = False


def test_get_run_by_id():
    with seven.TemporaryDirectory() as tmpdir_path:
        instance = DagsterInstance.from_ref(InstanceRef.from_dir(tmpdir_path))

        assert instance.get_runs() == []
        pipeline_run = PipelineRun("foo_pipeline", "new_run")
        assert instance.get_run_by_id(pipeline_run.run_id) is None

        instance._run_storage.add_run(pipeline_run)  # pylint: disable=protected-access

        assert instance.get_runs() == [pipeline_run]

        assert instance.get_run_by_id(pipeline_run.run_id) == pipeline_run

    # Run is created after we check whether it exists
    with seven.TemporaryDirectory() as tmpdir_path:
        instance = DagsterInstance.from_ref(InstanceRef.from_dir(tmpdir_path))
        run = PipelineRun(pipeline_name="foo_pipeline", run_id="bar_run")

        def _has_run(self, run_id):
            # This is uglier than we would like because there is no nonlocal keyword in py2
            global MOCK_HAS_RUN_CALLED  # pylint: disable=global-statement
            # pylint: disable=protected-access
            if not self._run_storage.has_run(run_id) and not MOCK_HAS_RUN_CALLED:
                self._run_storage.add_run(PipelineRun(pipeline_name="foo_pipeline", run_id=run_id))
                return False
            else:
                return self._run_storage.has_run(run_id)

        instance.has_run = types.MethodType(_has_run, instance)

        assert instance.get_run_by_id(run.run_id) is None

    # Run is created after we check whether it exists, but deleted before we can get it
    global MOCK_HAS_RUN_CALLED  # pylint:disable=global-statement
    MOCK_HAS_RUN_CALLED = False
    with seven.TemporaryDirectory() as tmpdir_path:
        instance = DagsterInstance.from_ref(InstanceRef.from_dir(tmpdir_path))
        run = PipelineRun(pipeline_name="foo_pipeline", run_id="bar_run")

        def _has_run(self, run_id):
            global MOCK_HAS_RUN_CALLED  # pylint: disable=global-statement
            # pylint: disable=protected-access
            if not self._run_storage.has_run(run_id) and not MOCK_HAS_RUN_CALLED:
                self._run_storage.add_run(PipelineRun(pipeline_name="foo_pipeline", run_id=run_id))
                MOCK_HAS_RUN_CALLED = True
                return False
            elif self._run_storage.has_run(run_id) and MOCK_HAS_RUN_CALLED:
                MOCK_HAS_RUN_CALLED = False
                return True
            else:
                return False

        instance.has_run = types.MethodType(_has_run, instance)
        assert instance.get_run_by_id(run.run_id) is None


def test_run_step_stats():
    @pipeline
    def simple():
        @solid
        def should_succeed(context):
            context.log.info("succeed")
            return "yay"

        @solid(input_defs=[InputDefinition("_input", str)], output_defs=[OutputDefinition(str)])
        def should_fail(context, _input):
            context.log.info("fail")
            raise Exception("booo")

        @solid
        def should_skip(context, _input):
            context.log.info("skip")
            return _input

        should_skip(should_fail(should_succeed()))

    with seven.TemporaryDirectory() as tmpdir_path:
        instance = DagsterInstance.from_ref(InstanceRef.from_dir(tmpdir_path))
        result = execute_pipeline(simple, instance=instance, raise_on_error=False)
        step_stats = sorted(instance.get_run_step_stats(result.run_id), key=lambda x: x.end_time)
        assert len(step_stats) == 3
        assert step_stats[0].step_key == "should_succeed.compute"
        assert step_stats[0].status == StepEventStatus.SUCCESS
        assert step_stats[0].end_time > step_stats[0].start_time
        assert step_stats[1].step_key == "should_fail.compute"
        assert step_stats[1].status == StepEventStatus.FAILURE
        assert step_stats[1].end_time > step_stats[0].start_time
        assert step_stats[2].step_key == "should_skip.compute"
        assert step_stats[2].status == StepEventStatus.SKIPPED
        assert step_stats[2].end_time > step_stats[0].start_time



import os
import time
from contextlib import contextmanager

import pytest

from dagster import DefaultRunLauncher, file_relative_path, pipeline, repository, solid
from dagster.core.definitions.reconstructable import ReconstructableRepository
from dagster.core.host_representation.handle import RepositoryLocationHandle
from dagster.core.host_representation.repository_location import (
    GrpcServerRepositoryLocation,
    InProcessRepositoryLocation,
)
from dagster.core.storage.pipeline_run import PipelineRunStatus
from dagster.core.test_utils import (
    instance_for_test,
    poll_for_event,
    poll_for_finished_run,
    poll_for_step_start,
)
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin
from dagster.grpc.server import GrpcServerProcess


@solid
def noop_solid(_):
    pass


@pipeline
def noop_pipeline():
    pass


@solid
def crashy_solid(_):
    os._exit(1)  # pylint: disable=W0212


@pipeline
def crashy_pipeline():
    crashy_solid()


@solid
def sleepy_solid(_):
    while True:
        time.sleep(0.1)


@pipeline
def sleepy_pipeline():
    sleepy_solid()


@solid
def return_one(_):
    return 1


@solid
def multiply_by_2(_, num):
    return num * 2


@solid
def multiply_by_3(_, num):
    return num * 3


@solid
def add(_, num1, num2):
    return num1 + num2


@pipeline
def math_diamond():
    one = return_one()
    add(multiply_by_2(one), multiply_by_3(one))


@repository
def nope():
    return [noop_pipeline, crashy_pipeline, sleepy_pipeline, math_diamond]


def test_repo_construction():
    repo_yaml = file_relative_path(__file__, "repo.yaml")
    assert ReconstructableRepository.from_legacy_repository_yaml(repo_yaml).get_definition()


@contextmanager
def get_external_pipeline_from_grpc_server_repository(pipeline_name):
    repo_yaml = file_relative_path(__file__, "repo.yaml")
    recon_repo = ReconstructableRepository.from_legacy_repository_yaml(repo_yaml)
    loadable_target_origin = recon_repo.get_origin().loadable_target_origin
    server_process = GrpcServerProcess(loadable_target_origin=loadable_target_origin)

    try:
        with server_process.create_ephemeral_client() as api_client:
            repository_location = GrpcServerRepositoryLocation(
                RepositoryLocationHandle.create_grpc_server_location(
                    location_name="test",
                    port=api_client.port,
                    socket=api_client.socket,
                    host=api_client.host,
                )
            )

            yield repository_location.get_repository("nope").get_full_external_pipeline(
                pipeline_name
            )
    finally:
        server_process.wait()


@contextmanager
def get_external_pipeline_from_managed_grpc_python_env_repository(pipeline_name):
    repository_location_handle = RepositoryLocationHandle.create_process_bound_grpc_server_location(
        loadable_target_origin=LoadableTargetOrigin(
            attribute="nope",
            python_file=file_relative_path(__file__, "test_cli_api_run_launcher.py"),
        ),
        location_name="nope",
    )
    repository_location = GrpcServerRepositoryLocation(repository_location_handle)
    try:
        yield repository_location.get_repository("nope").get_full_external_pipeline(pipeline_name)
    finally:
        repository_location_handle.cleanup()


@contextmanager
def get_external_pipeline_from_in_process_location(pipeline_name):
    repo_yaml = file_relative_path(__file__, "repo.yaml")
    recon_repo = ReconstructableRepository.from_legacy_repository_yaml(repo_yaml)
    yield (
        InProcessRepositoryLocation(recon_repo)
        .get_repository("nope")
        .get_full_external_pipeline(pipeline_name)
    )


@pytest.mark.parametrize(
    "get_external_pipeline",
    [
        get_external_pipeline_from_grpc_server_repository,
        get_external_pipeline_from_managed_grpc_python_env_repository,
        get_external_pipeline_from_in_process_location,
    ],
)
def test_successful_run(get_external_pipeline):  # pylint: disable=redefined-outer-name
    with instance_for_test() as instance:
        pipeline_run = instance.create_run_for_pipeline(pipeline_def=noop_pipeline, run_config=None)

        with get_external_pipeline(pipeline_run.pipeline_name) as external_pipeline:
            run_id = pipeline_run.run_id

            assert instance.get_run_by_id(run_id).status == PipelineRunStatus.NOT_STARTED

            launcher = instance.run_launcher
            launcher.launch_run(
                instance=instance, run=pipeline_run, external_pipeline=external_pipeline
            )

            pipeline_run = instance.get_run_by_id(run_id)
            assert pipeline_run
            assert pipeline_run.run_id == run_id

            pipeline_run = poll_for_finished_run(instance, run_id)
            assert pipeline_run.status == PipelineRunStatus.SUCCESS


@pytest.mark.parametrize(
    "get_external_pipeline",
    [
        get_external_pipeline_from_grpc_server_repository,
        get_external_pipeline_from_managed_grpc_python_env_repository,
        get_external_pipeline_from_in_process_location,
    ],
)
def test_crashy_run(get_external_pipeline):  # pylint: disable=redefined-outer-name
    with instance_for_test() as instance:
        pipeline_run = instance.create_run_for_pipeline(
            pipeline_def=crashy_pipeline, run_config=None
        )

        with get_external_pipeline(pipeline_run.pipeline_name) as external_pipeline:

            run_id = pipeline_run.run_id

            assert instance.get_run_by_id(run_id).status == PipelineRunStatus.NOT_STARTED

            launcher = instance.run_launcher
            launcher.launch_run(instance, pipeline_run, external_pipeline)

            failed_pipeline_run = instance.get_run_by_id(run_id)

            assert failed_pipeline_run
            assert failed_pipeline_run.run_id == run_id

            failed_pipeline_run = poll_for_finished_run(instance, run_id, timeout=5)
            assert failed_pipeline_run.status == PipelineRunStatus.FAILURE

            event_records = instance.all_logs(run_id)

            message = "Pipeline execution process for {run_id} unexpectedly exited.".format(
                run_id=run_id
            )

            assert _message_exists(event_records, message)


@pytest.mark.parametrize(
    "get_external_pipeline,in_process",
    [
        (get_external_pipeline_from_grpc_server_repository, False),
        (get_external_pipeline_from_managed_grpc_python_env_repository, False),
        (get_external_pipeline_from_in_process_location, True),
    ],
)
def test_terminated_run(get_external_pipeline, in_process):  # pylint: disable=redefined-outer-name
    with instance_for_test() as instance:
        pipeline_run = instance.create_run_for_pipeline(
            pipeline_def=sleepy_pipeline, run_config=None
        )

        with get_external_pipeline(pipeline_run.pipeline_name) as external_pipeline:
            run_id = pipeline_run.run_id

            assert instance.get_run_by_id(run_id).status == PipelineRunStatus.NOT_STARTED

            launcher = instance.run_launcher
            launcher.launch_run(instance, pipeline_run, external_pipeline)

            poll_for_step_start(instance, run_id)

            assert launcher.can_terminate(run_id)
            assert launcher.terminate(run_id)

            terminated_pipeline_run = poll_for_finished_run(instance, run_id, timeout=30)
            terminated_pipeline_run = instance.get_run_by_id(run_id)
            assert terminated_pipeline_run.status == PipelineRunStatus.FAILURE

            poll_for_event(
                instance, run_id, event_type="ENGINE_EVENT", message="Process for pipeline exited"
            )

            run_logs = instance.all_logs(run_id)
            event_types = [event.dagster_event.event_type_value for event in run_logs]
            if in_process:
                assert event_types == [
                    "ENGINE_EVENT",
                    "ENGINE_EVENT",
                    "PIPELINE_START",
                    "ENGINE_EVENT",
                    "STEP_START",
                    "STEP_FAILURE",
                    "PIPELINE_FAILURE",
                    "ENGINE_EVENT",
                ]
            else:
                assert event_types == [
                    "ENGINE_EVENT",
                    "PIPELINE_START",
                    "ENGINE_EVENT",
                    "STEP_START",
                    "STEP_FAILURE",
                    "PIPELINE_FAILURE",
                    "ENGINE_EVENT",
                    "ENGINE_EVENT",
                ]


def _get_engine_events(event_records):
    for er in event_records:
        if er.dagster_event and er.dagster_event.is_engine_event:
            yield er


def _get_successful_step_keys(event_records):

    step_keys = set()

    for er in event_records:
        if er.dagster_event and er.dagster_event.is_step_success:
            step_keys.add(er.dagster_event.step_key)

    return step_keys


def _message_exists(event_records, message_text):
    for event_record in event_records:
        if message_text in event_record.message:
            return True

    return False


@pytest.mark.parametrize(
    "get_external_pipeline",
    [
        get_external_pipeline_from_grpc_server_repository,
        get_external_pipeline_from_managed_grpc_python_env_repository,
        get_external_pipeline_from_in_process_location,
    ],
)
def test_single_solid_selection_execution(
    get_external_pipeline,
):  # pylint: disable=redefined-outer-name
    with instance_for_test() as instance:
        pipeline_run = instance.create_run_for_pipeline(
            pipeline_def=math_diamond, run_config=None, solids_to_execute={"return_one"}
        )
        run_id = pipeline_run.run_id

        assert instance.get_run_by_id(run_id).status == PipelineRunStatus.NOT_STARTED

        with get_external_pipeline(pipeline_run.pipeline_name) as external_pipeline:
            launcher = instance.run_launcher
            launcher.launch_run(instance, pipeline_run, external_pipeline)
            finished_pipeline_run = poll_for_finished_run(instance, run_id)

            event_records = instance.all_logs(run_id)

            assert finished_pipeline_run
            assert finished_pipeline_run.run_id == run_id
            assert finished_pipeline_run.status == PipelineRunStatus.SUCCESS

            assert _get_successful_step_keys(event_records) == {"return_one.compute"}


@pytest.mark.parametrize(
    "get_external_pipeline",
    [
        get_external_pipeline_from_grpc_server_repository,
        get_external_pipeline_from_managed_grpc_python_env_repository,
        get_external_pipeline_from_in_process_location,
    ],
)
def test_multi_solid_selection_execution(
    get_external_pipeline,
):  # pylint: disable=redefined-outer-name
    with instance_for_test() as instance:
        pipeline_run = instance.create_run_for_pipeline(
            pipeline_def=math_diamond,
            run_config=None,
            solids_to_execute={"return_one", "multiply_by_2"},
        )
        run_id = pipeline_run.run_id

        assert instance.get_run_by_id(run_id).status == PipelineRunStatus.NOT_STARTED

        with get_external_pipeline(pipeline_run.pipeline_name) as external_pipeline:
            launcher = instance.run_launcher
            launcher.launch_run(instance, pipeline_run, external_pipeline)
            finished_pipeline_run = poll_for_finished_run(instance, run_id)

            event_records = instance.all_logs(run_id)

            assert finished_pipeline_run
            assert finished_pipeline_run.run_id == run_id
            assert finished_pipeline_run.status == PipelineRunStatus.SUCCESS

            assert _get_successful_step_keys(event_records) == {
                "return_one.compute",
                "multiply_by_2.compute",
            }


@pytest.mark.parametrize(
    "get_external_pipeline,in_process",
    [
        (get_external_pipeline_from_grpc_server_repository, False),
        (get_external_pipeline_from_managed_grpc_python_env_repository, False),
        (get_external_pipeline_from_in_process_location, True),
    ],
)
def test_engine_events(get_external_pipeline, in_process):  # pylint: disable=redefined-outer-name
    with instance_for_test() as instance:
        pipeline_run = instance.create_run_for_pipeline(pipeline_def=math_diamond, run_config=None)
        run_id = pipeline_run.run_id

        assert instance.get_run_by_id(run_id).status == PipelineRunStatus.NOT_STARTED

        with get_external_pipeline(pipeline_run.pipeline_name) as external_pipeline:
            launcher = instance.run_launcher
            launcher.launch_run(instance, pipeline_run, external_pipeline)
            finished_pipeline_run = poll_for_finished_run(instance, run_id)

            assert finished_pipeline_run
            assert finished_pipeline_run.run_id == run_id
            assert finished_pipeline_run.status == PipelineRunStatus.SUCCESS

            poll_for_event(
                instance, run_id, event_type="ENGINE_EVENT", message="Process for pipeline exited"
            )
            event_records = instance.all_logs(run_id)

            if in_process:
                (
                    about_to_start,
                    started_process,
                    executing_steps,
                    finished_steps,
                    process_exited,
                ) = tuple(_get_engine_events(event_records))

                assert "About to start process" in about_to_start.message
                assert "Started process for pipeline" in started_process.message
                assert "Executing steps in process" in executing_steps.message
                assert "Finished steps in process" in finished_steps.message
                assert "Process for pipeline exited" in process_exited.message
            else:
                (started_process, executing_steps, finished_steps, process_exited) = tuple(
                    _get_engine_events(event_records)
                )

                assert "Started process for pipeline" in started_process.message
                assert "Executing steps in process" in executing_steps.message
                assert "Finished steps in process" in finished_steps.message
                assert "Process for pipeline exited" in process_exited.message


def test_not_initialized():  # pylint: disable=redefined-outer-name
    run_launcher = DefaultRunLauncher()
    run_id = "dummy"

    assert run_launcher.join() is None
    assert run_launcher.can_terminate(run_id) is False
    assert run_launcher.terminate(run_id) is False



import pytest

from dagster import seven
from dagster.api.execute_run import cli_api_execute_run, sync_execute_run_grpc
from dagster.core.instance import DagsterInstance
from dagster.grpc.server import GrpcServerProcess

from .utils import (
    get_foo_grpc_pipeline_handle,
    get_foo_pipeline_handle,
    legacy_get_foo_pipeline_handle,
)


@pytest.mark.parametrize(
    "pipeline_handle", [get_foo_pipeline_handle(), legacy_get_foo_pipeline_handle()],
)
def test_execute_run_api(pipeline_handle):
    with seven.TemporaryDirectory() as temp_dir:
        instance = DagsterInstance.local_temp(temp_dir)
        pipeline_run = instance.create_run(
            pipeline_name="foo",
            run_id=None,
            run_config={},
            mode="default",
            solids_to_execute=None,
            step_keys_to_execute=None,
            status=None,
            tags=None,
            root_run_id=None,
            parent_run_id=None,
            pipeline_snapshot=None,
            execution_plan_snapshot=None,
            parent_pipeline_snapshot=None,
        )
        events = cli_api_execute_run(
            instance=instance,
            pipeline_origin=pipeline_handle.get_origin(),
            pipeline_run=pipeline_run,
        )

    assert len(events) == 11
    assert [event.event_type_value for event in events] == [
        "PIPELINE_START",
        "ENGINE_EVENT",
        "STEP_START",
        "STEP_OUTPUT",
        "STEP_SUCCESS",
        "STEP_START",
        "STEP_INPUT",
        "STEP_OUTPUT",
        "STEP_SUCCESS",
        "ENGINE_EVENT",
        "PIPELINE_SUCCESS",
    ]


def test_execute_run_api_grpc_server_handle():
    with seven.TemporaryDirectory() as temp_dir:
        instance = DagsterInstance.local_temp(temp_dir)
        with get_foo_grpc_pipeline_handle() as pipeline_handle:
            pipeline_run = instance.create_run(
                pipeline_name="foo",
                run_id=None,
                run_config={},
                mode="default",
                solids_to_execute=None,
                step_keys_to_execute=None,
                status=None,
                tags=None,
                root_run_id=None,
                parent_run_id=None,
                pipeline_snapshot=None,
                execution_plan_snapshot=None,
                parent_pipeline_snapshot=None,
            )
            events = [
                event
                for event in sync_execute_run_grpc(
                    api_client=pipeline_handle.repository_handle.repository_location_handle.client,
                    instance_ref=instance.get_ref(),
                    pipeline_origin=pipeline_handle.get_origin(),
                    pipeline_run=pipeline_run,
                )
            ]

    assert len(events) == 14
    assert [event.event_type_value for event in events] == [
        "ENGINE_EVENT",
        "ENGINE_EVENT",
        "PIPELINE_START",
        "ENGINE_EVENT",
        "STEP_START",
        "STEP_OUTPUT",
        "STEP_SUCCESS",
        "STEP_START",
        "STEP_INPUT",
        "STEP_OUTPUT",
        "STEP_SUCCESS",
        "ENGINE_EVENT",
        "PIPELINE_SUCCESS",
        "ENGINE_EVENT",
    ]


@pytest.mark.parametrize(
    "pipeline_handle", [get_foo_pipeline_handle()],
)
def test_execute_run_api_grpc_python_handle(pipeline_handle):
    with seven.TemporaryDirectory() as temp_dir:
        instance = DagsterInstance.local_temp(temp_dir)
        pipeline_run = instance.create_run(
            pipeline_name="foo",
            run_id=None,
            run_config={},
            mode="default",
            solids_to_execute=None,
            step_keys_to_execute=None,
            status=None,
            tags=None,
            root_run_id=None,
            parent_run_id=None,
            pipeline_snapshot=None,
            execution_plan_snapshot=None,
            parent_pipeline_snapshot=None,
        )

        loadable_target_origin = (
            pipeline_handle.get_origin().repository_origin.loadable_target_origin
        )

        with GrpcServerProcess(
            loadable_target_origin, max_workers=2
        ).create_ephemeral_client() as api_client:
            events = [
                event
                for event in sync_execute_run_grpc(
                    api_client=api_client,
                    instance_ref=instance.get_ref(),
                    pipeline_origin=pipeline_handle.get_origin(),
                    pipeline_run=pipeline_run,
                )
            ]

            assert len(events) == 14
            assert [event.event_type_value for event in events] == [
                "ENGINE_EVENT",
                "ENGINE_EVENT",
                "PIPELINE_START",
                "ENGINE_EVENT",
                "STEP_START",
                "STEP_OUTPUT",
                "STEP_SUCCESS",
                "STEP_START",
                "STEP_INPUT",
                "STEP_OUTPUT",
                "STEP_SUCCESS",
                "ENGINE_EVENT",
                "PIPELINE_SUCCESS",
                "ENGINE_EVENT",
            ]



from abc import abstractmethod

import six
import sqlalchemy as db

from dagster import check
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.scheduler import ScheduleState, ScheduleTick
from dagster.core.scheduler.scheduler import (
    ScheduleTickData,
    ScheduleTickStatsSnapshot,
    ScheduleTickStatus,
)
from dagster.serdes import deserialize_json_to_dagster_namedtuple, serialize_dagster_namedtuple
from dagster.utils import utc_datetime_from_timestamp

from .base import ScheduleStorage
from .schema import ScheduleTable, ScheduleTickTable


class SqlScheduleStorage(ScheduleStorage):
    """Base class for SQL backed schedule storage
    """

    @abstractmethod
    def connect(self):
        """Context manager yielding a sqlalchemy.engine.Connection."""

    def execute(self, query):
        with self.connect() as conn:
            result_proxy = conn.execute(query)
            res = result_proxy.fetchall()
            result_proxy.close()

        return res

    def _deserialize_rows(self, rows):
        return list(map(lambda r: deserialize_json_to_dagster_namedtuple(r[0]), rows))

    def all_stored_schedule_state(self, repository_origin_id=None):
        base_query = db.select(
            [ScheduleTable.c.schedule_body, ScheduleTable.c.schedule_origin_id]
        ).select_from(ScheduleTable)

        if repository_origin_id:
            query = base_query.where(ScheduleTable.c.repository_origin_id == repository_origin_id)
        else:
            query = base_query

        rows = self.execute(query)
        return self._deserialize_rows(rows)

    def get_schedule_state(self, schedule_origin_id):
        check.str_param(schedule_origin_id, "schedule_origin_id")

        query = (
            db.select([ScheduleTable.c.schedule_body])
            .select_from(ScheduleTable)
            .where(ScheduleTable.c.schedule_origin_id == schedule_origin_id)
        )

        rows = self.execute(query)
        return deserialize_json_to_dagster_namedtuple(rows[0][0]) if len(rows) else None

    def get_schedule_ticks(self, schedule_origin_id):
        check.str_param(schedule_origin_id, "schedule_origin_id")

        query = (
            db.select([ScheduleTickTable.c.id, ScheduleTickTable.c.tick_body])
            .select_from(ScheduleTickTable)
            .where(ScheduleTickTable.c.schedule_origin_id == schedule_origin_id)
            .order_by(ScheduleTickTable.c.id.desc())
        )

        rows = self.execute(query)
        return list(
            map(lambda r: ScheduleTick(r[0], deserialize_json_to_dagster_namedtuple(r[1])), rows)
        )

    def get_schedule_tick_stats(self, schedule_origin_id):
        check.str_param(schedule_origin_id, "schedule_origin_id")

        query = (
            db.select([ScheduleTickTable.c.status, db.func.count()])
            .select_from(ScheduleTickTable)
            .where(ScheduleTickTable.c.schedule_origin_id == schedule_origin_id)
            .group_by(ScheduleTickTable.c.status)
        )

        rows = self.execute(query)

        counts = {}
        for status, count in rows:
            counts[status] = count

        return ScheduleTickStatsSnapshot(
            ticks_started=counts.get(ScheduleTickStatus.STARTED.value, 0),
            ticks_succeeded=counts.get(ScheduleTickStatus.SUCCESS.value, 0),
            ticks_skipped=counts.get(ScheduleTickStatus.SKIPPED.value, 0),
            ticks_failed=counts.get(ScheduleTickStatus.FAILURE.value, 0),
        )

    def create_schedule_tick(self, schedule_tick_data):
        check.inst_param(schedule_tick_data, "schedule_tick_data", ScheduleTickData)

        with self.connect() as conn:
            try:
                tick_insert = ScheduleTickTable.insert().values(  # pylint: disable=no-value-for-parameter
                    schedule_origin_id=schedule_tick_data.schedule_origin_id,
                    status=schedule_tick_data.status.value,
                    timestamp=utc_datetime_from_timestamp(schedule_tick_data.timestamp),
                    tick_body=serialize_dagster_namedtuple(schedule_tick_data),
                )
                result = conn.execute(tick_insert)
                tick_id = result.inserted_primary_key[0]
                return ScheduleTick(tick_id, schedule_tick_data)
            except db.exc.IntegrityError as exc:
                six.raise_from(
                    DagsterInvariantViolationError(
                        "Unable to insert ScheduleTick for schedule {schedule_name} in storage".format(
                            schedule_name=schedule_tick_data.schedule_name,
                        )
                    ),
                    exc,
                )

    def update_schedule_tick(self, tick):
        check.inst_param(tick, "tick", ScheduleTick)

        with self.connect() as conn:
            conn.execute(
                ScheduleTickTable.update()  # pylint: disable=no-value-for-parameter
                .where(ScheduleTickTable.c.id == tick.tick_id)
                .values(
                    status=tick.status.value,
                    tick_body=serialize_dagster_namedtuple(tick.schedule_tick_data),
                )
            )

        return tick

    def add_schedule_state(self, schedule):
        check.inst_param(schedule, "schedule", ScheduleState)
        with self.connect() as conn:
            try:
                schedule_insert = ScheduleTable.insert().values(  # pylint: disable=no-value-for-parameter
                    schedule_origin_id=schedule.schedule_origin_id,
                    repository_origin_id=schedule.repository_origin_id,
                    status=schedule.status.value,
                    schedule_body=serialize_dagster_namedtuple(schedule),
                )
                conn.execute(schedule_insert)
            except db.exc.IntegrityError as exc:
                six.raise_from(
                    DagsterInvariantViolationError(
                        "ScheduleState {id} is already present "
                        "in storage".format(id=schedule.schedule_origin_id,)
                    ),
                    exc,
                )

        return schedule

    def update_schedule_state(self, schedule):
        check.inst_param(schedule, "schedule", ScheduleState)
        if not self.get_schedule_state(schedule.schedule_origin_id):
            raise DagsterInvariantViolationError(
                "ScheduleState {id} is not present in storage".format(
                    id=schedule.schedule_origin_id
                )
            )

        with self.connect() as conn:
            conn.execute(
                ScheduleTable.update()  # pylint: disable=no-value-for-parameter
                .where(ScheduleTable.c.schedule_origin_id == schedule.schedule_origin_id)
                .values(
                    status=schedule.status.value,
                    schedule_body=serialize_dagster_namedtuple(schedule),
                )
            )

    def delete_schedule_state(self, schedule_origin_id):
        check.str_param(schedule_origin_id, "schedule_origin_id")

        if not self.get_schedule_state(schedule_origin_id):
            raise DagsterInvariantViolationError(
                "ScheduleState {id} is not present in storage".format(id=schedule_origin_id)
            )

        with self.connect() as conn:
            conn.execute(
                ScheduleTable.delete().where(  # pylint: disable=no-value-for-parameter
                    ScheduleTable.c.schedule_origin_id == schedule_origin_id
                )
            )

    def wipe(self):
        """Clears the schedule storage."""
        with self.connect() as conn:
            # https://stackoverflow.com/a/54386260/324449
            conn.execute(ScheduleTable.delete())  # pylint: disable=no-value-for-parameter
            conn.execute(ScheduleTickTable.delete())  # pylint: disable=no-value-for-parameter



import subprocess
import sys

import pytest


@pytest.mark.skipif(sys.version_info < (3, 6), reason="black is not available for python < 3.6")
def test_compiled_protobuf():
    expected_status = subprocess.check_output(["git", "status", "--porcelain"])
    subprocess.check_output([sys.executable, "-m", "dagster.grpc.compile"])
    assert expected_status == subprocess.check_output(["git", "status", "--porcelain"])



import os

import pytest


def aws_credentials_present():
    return os.getenv("AWS_ACCESS_KEY_ID") and os.getenv("AWS_SECRET_ACCESS_KEY")


aws = pytest.mark.skipif(not aws_credentials_present(), reason="Couldn't find AWS credentials")

nettest = pytest.mark.nettest



import sys
from collections import defaultdict
from contextlib import contextmanager

import pytest

from dagster import check
from dagster.check import (
    CheckError,
    ElementCheckError,
    NotImplementedCheckError,
    ParameterCheckError,
)
from dagster.utils import frozendict, frozenlist


def test_int_param():
    assert check.int_param(-1, "param_name") == -1
    assert check.int_param(0, "param_name") == 0
    assert check.int_param(1, "param_name") == 1

    with pytest.raises(ParameterCheckError):
        check.int_param(None, "param_name")

    with pytest.raises(ParameterCheckError):
        check.int_param("s", "param_name")


def test_int_value_param():
    assert check.int_value_param(-1, -1, "param_name") == -1
    with pytest.raises(ParameterCheckError):
        check.int_value_param(None, -1, "param_name")

    with pytest.raises(ParameterCheckError):
        check.int_value_param(1, 0, "param_name")


def test_opt_int_param():
    assert check.opt_int_param(-1, "param_name") == -1
    assert check.opt_int_param(0, "param_name") == 0
    assert check.opt_int_param(1, "param_name") == 1
    assert check.opt_int_param(None, "param_name") is None

    with pytest.raises(ParameterCheckError):
        check.opt_int_param("s", "param_name")


def test_float_param():
    assert check.float_param(-1.0, "param_name") == -1.0
    assert check.float_param(0.0, "param_name") == 0.0
    assert check.float_param(1.1, "param_name") == 1.1

    with pytest.raises(ParameterCheckError):
        check.float_param(None, "param_name")

    with pytest.raises(ParameterCheckError):
        check.float_param("s", "param_name")

    with pytest.raises(ParameterCheckError):
        check.float_param(1, "param_name")

    with pytest.raises(ParameterCheckError):
        check.float_param(0, "param_name")


def test_opt_float_param():
    assert check.opt_float_param(-1.0, "param_name") == -1.0
    assert check.opt_float_param(0.0, "param_name") == 0.0
    assert check.opt_float_param(1.1, "param_name") == 1.1
    assert check.opt_float_param(None, "param_name") is None

    with pytest.raises(ParameterCheckError):
        check.opt_float_param("s", "param_name")


def test_list_param():
    assert check.list_param([], "list_param") == []
    assert check.list_param(frozenlist(), "list_param") == []

    with pytest.raises(ParameterCheckError):
        check.list_param(None, "list_param")

    with pytest.raises(ParameterCheckError):
        check.list_param("3u4", "list_param")


def test_set_param():
    assert check.set_param(set(), "set_param") == set()
    assert check.set_param(frozenset(), "set_param") == set()

    with pytest.raises(ParameterCheckError):
        check.set_param(None, "set_param")

    with pytest.raises(ParameterCheckError):
        check.set_param("3u4", "set_param")

    obj_set = {1}
    assert check.set_param(obj_set, "set_param") == obj_set

    obj_set_two = {1, 1, 2}
    obj_set_two_deduped = {1, 2}
    assert check.set_param(obj_set_two, "set_param") == obj_set_two_deduped
    assert check.set_param(obj_set_two, "set_param", of_type=int) == obj_set_two_deduped

    with pytest.raises(CheckError, match="Did you pass a class"):
        check.set_param({str}, "set_param", of_type=int)

    with pytest.raises(CheckError, match="Member of set mismatches type"):
        check.set_param({"foo"}, "set_param", of_type=int)


def test_typed_list_param():
    class Foo(object):
        pass

    class Bar(object):
        pass

    assert check.list_param([], "list_param", Foo) == []
    foo_list = [Foo()]
    assert check.list_param(foo_list, "list_param", Foo) == foo_list

    with pytest.raises(CheckError):
        check.list_param([Bar()], "list_param", Foo)

    with pytest.raises(CheckError):
        check.list_param([None], "list_param", Foo)


def test_is_tuple():
    assert check.is_tuple(()) == ()

    with pytest.raises(CheckError):
        check.is_tuple(None)

    with pytest.raises(CheckError):
        check.is_tuple("3u4")

    with pytest.raises(CheckError, match="Did you pass a class"):
        check.is_tuple((str,), of_type=int)


def test_typed_is_tuple():
    class Foo(object):
        pass

    class Bar(object):
        pass

    assert check.is_tuple((), Foo) == ()
    foo_tuple = (Foo(),)
    assert check.is_tuple(foo_tuple, Foo) == foo_tuple

    with pytest.raises(CheckError):
        check.is_tuple((Bar(),), Foo)

    with pytest.raises(CheckError):
        check.is_tuple((None,), Foo)


def test_typed_is_list():
    class Foo(object):
        pass

    class Bar(object):
        pass

    assert check.is_list([], Foo) == []
    foo_list = [Foo()]
    assert check.is_list(foo_list, Foo) == foo_list

    with pytest.raises(CheckError):
        check.is_list([Bar()], Foo)

    with pytest.raises(CheckError):
        check.is_list([None], Foo)


def test_opt_list_param():
    assert check.opt_list_param(None, "list_param") == []
    assert check.opt_list_param([], "list_param") == []
    assert check.opt_list_param(frozenlist(), "list_param") == []
    obj_list = [1]
    assert check.list_param(obj_list, "list_param") == obj_list

    with pytest.raises(ParameterCheckError):
        check.opt_list_param(0, "list_param")

    with pytest.raises(ParameterCheckError):
        check.opt_list_param("", "list_param")

    with pytest.raises(ParameterCheckError):
        check.opt_list_param("3u4", "list_param")


def test_opt_set_param():
    assert check.opt_set_param(None, "set_param") == set()
    assert check.opt_set_param(set(), "set_param") == set()
    assert check.opt_set_param(frozenset(), "set_param") == set()
    assert check.opt_set_param({3}, "set_param") == {3}

    with pytest.raises(ParameterCheckError):
        check.opt_set_param(0, "set_param")

    with pytest.raises(ParameterCheckError):
        check.opt_set_param("", "set_param")

    with pytest.raises(ParameterCheckError):
        check.opt_set_param("3u4", "set_param")


def test_opt_nullable_list_param():
    assert check.opt_nullable_list_param(None, "list_param") is None
    assert check.opt_nullable_list_param([], "list_param") == []
    assert check.opt_nullable_list_param(frozenlist(), "list_param") == []
    obj_list = [1]
    assert check.opt_nullable_list_param(obj_list, "list_param") == obj_list

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_list_param(0, "list_param")

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_list_param("", "list_param")

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_list_param("3u4", "list_param")


def test_opt_typed_list_param():
    class Foo(object):
        pass

    class Bar(object):
        pass

    assert check.opt_list_param(None, "list_param", Foo) == []
    assert check.opt_list_param([], "list_param", Foo) == []
    foo_list = [Foo()]
    assert check.opt_list_param(foo_list, "list_param", Foo) == foo_list

    with pytest.raises(CheckError):
        check.opt_list_param([Bar()], "list_param", Foo)

    with pytest.raises(CheckError):
        check.opt_list_param([None], "list_param", Foo)


def test_dict_param():
    assert check.dict_param({}, "dict_param") == {}
    assert check.dict_param(frozendict(), "dict_param") == {}
    ddict = {"a": 2}
    assert check.dict_param(ddict, "dict_param") == ddict

    with pytest.raises(ParameterCheckError):
        check.dict_param(None, "dict_param")

    with pytest.raises(ParameterCheckError):
        check.dict_param(0, "dict_param")

    with pytest.raises(ParameterCheckError):
        check.dict_param(1, "dict_param")

    with pytest.raises(ParameterCheckError):
        check.dict_param("foo", "dict_param")

    with pytest.raises(ParameterCheckError):
        check.dict_param(["foo"], "dict_param")

    with pytest.raises(ParameterCheckError):
        check.dict_param([], "dict_param")


def test_dict_param_with_type():
    str_to_int = {"str": 1}
    assert check.dict_param(str_to_int, "str_to_int", key_type=str, value_type=int)
    assert check.dict_param(str_to_int, "str_to_int", value_type=int)
    assert check.dict_param(str_to_int, "str_to_int", key_type=str)
    assert check.dict_param(str_to_int, "str_to_int")

    assert check.dict_param({}, "str_to_int", key_type=str, value_type=int) == {}
    assert check.dict_param({}, "str_to_int", value_type=int) == {}
    assert check.dict_param({}, "str_to_int", key_type=str) == {}
    assert check.dict_param({}, "str_to_int") == {}

    class Wrong(object):
        pass

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, "str_to_int", key_type=Wrong, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, "str_to_int", key_type=Wrong, value_type=int)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, "str_to_int", key_type=str, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, "str_to_int", key_type=Wrong)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, "str_to_int", value_type=Wrong)


def test_opt_dict_param_with_type():
    str_to_int = {"str": 1}
    assert check.opt_dict_param(str_to_int, "str_to_int", key_type=str, value_type=int)
    assert check.opt_dict_param(str_to_int, "str_to_int", value_type=int)
    assert check.opt_dict_param(str_to_int, "str_to_int", key_type=str)
    assert check.opt_dict_param(str_to_int, "str_to_int")

    assert check.opt_dict_param({}, "str_to_int", key_type=str, value_type=int) == {}
    assert check.opt_dict_param({}, "str_to_int", value_type=int) == {}
    assert check.opt_dict_param({}, "str_to_int", key_type=str) == {}
    assert check.opt_dict_param({}, "str_to_int") == {}

    assert check.opt_dict_param(None, "str_to_int", key_type=str, value_type=int) == {}
    assert check.opt_dict_param(None, "str_to_int", value_type=int) == {}
    assert check.opt_dict_param(None, "str_to_int", key_type=str) == {}
    assert check.opt_dict_param(None, "str_to_int") == {}

    class Wrong(object):
        pass

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, "str_to_int", key_type=Wrong, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, "str_to_int", key_type=Wrong, value_type=int)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, "str_to_int", key_type=str, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, "str_to_int", key_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, "str_to_int", value_type=Wrong)


def test_opt_dict_param():
    assert check.opt_dict_param(None, "opt_dict_param") == {}
    assert check.opt_dict_param({}, "opt_dict_param") == {}
    assert check.opt_dict_param(frozendict(), "opt_dict_param") == {}
    ddict = {"a": 2}
    assert check.opt_dict_param(ddict, "opt_dict_param") == ddict

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(0, "opt_dict_param")

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(1, "opt_dict_param")

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param("foo", "opt_dict_param")

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(["foo"], "opt_dict_param")

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param([], "opt_dict_param")


def test_opt_nullable_dict_param():
    assert check.opt_nullable_dict_param(None, "opt_nullable_dict_param") is None
    assert check.opt_nullable_dict_param({}, "opt_nullable_dict_param") == {}
    assert check.opt_nullable_dict_param(frozendict(), "opt_nullable_dict_param") == {}
    ddict = {"a": 2}
    assert check.opt_nullable_dict_param(ddict, "opt_nullable_dict_param") == ddict

    class Foo(object):
        pass

    class Bar(Foo):
        pass

    ddict_class = {"a": Bar}
    assert (
        check.opt_nullable_dict_param(ddict_class, "opt_nullable_dict_param", value_class=Foo)
        == ddict_class
    )

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_dict_param(1, "opt_nullable_dict_param")

    with pytest.raises(ParameterCheckError):
        check.opt_nullable_dict_param("foo", "opt_nullable_dict_param")


def test_str_param():
    assert check.str_param("a", "str_param") == "a"
    assert check.str_param("", "str_param") == ""
    assert check.str_param(u"a", "unicode_param") == u"a"

    with pytest.raises(ParameterCheckError):
        check.str_param(None, "str_param")

    with pytest.raises(ParameterCheckError):
        check.str_param(0, "str_param")

    with pytest.raises(ParameterCheckError):
        check.str_param(1, "str_param")


def test_opt_str_param():
    assert check.opt_str_param("a", "str_param") == "a"
    assert check.opt_str_param("", "str_param") == ""
    assert check.opt_str_param(u"a", "unicode_param") == u"a"
    assert check.opt_str_param(None, "str_param") is None
    assert check.opt_str_param(None, "str_param", "foo") == "foo"

    with pytest.raises(ParameterCheckError):
        check.opt_str_param(0, "str_param")

    with pytest.raises(ParameterCheckError):
        check.opt_str_param(1, "str_param")


def test_opt_nonempty_str_param():
    assert check.opt_nonempty_str_param("a", "str_param") == "a"
    assert check.opt_nonempty_str_param("", "str_param") is None
    assert check.opt_nonempty_str_param("", "str_param", "foo") == "foo"
    assert check.opt_nonempty_str_param(u"a", "unicode_param") == u"a"
    assert check.opt_nonempty_str_param(None, "str_param") is None
    assert check.opt_nonempty_str_param(None, "str_param", "foo") == "foo"

    with pytest.raises(ParameterCheckError):
        check.opt_nonempty_str_param(0, "str_param")

    with pytest.raises(ParameterCheckError):
        check.opt_nonempty_str_param(1, "str_param")


def test_bool_param():
    assert check.bool_param(True, "b") is True
    assert check.bool_param(False, "b") is False

    with pytest.raises(ParameterCheckError):
        check.bool_param(None, "b")

    with pytest.raises(ParameterCheckError):
        check.bool_param(0, "b")

    with pytest.raises(ParameterCheckError):
        check.bool_param("val", "b")


def test_opt_bool_param():
    assert check.opt_bool_param(True, "b") is True
    assert check.opt_bool_param(False, "b") is False
    assert check.opt_bool_param(None, "b") is None
    assert check.opt_bool_param(None, "b", True) is True
    assert check.opt_bool_param(None, "b", False) is False

    with pytest.raises(ParameterCheckError):
        check.opt_bool_param(0, "b")

    with pytest.raises(ParameterCheckError):
        check.opt_bool_param("val", "b")


def test_callable_param():
    lamb = lambda: 1
    assert check.callable_param(lamb, "lamb") == lamb

    with pytest.raises(ParameterCheckError):
        check.callable_param(None, "lamb")

    with pytest.raises(ParameterCheckError):
        check.callable_param(2, "lamb")


def test_opt_callable_param():
    lamb = lambda: 1
    assert check.opt_callable_param(lamb, "lamb") == lamb
    assert check.opt_callable_param(None, "lamb") is None
    assert check.opt_callable_param(None, "lamb", default=None) is None
    assert check.opt_callable_param(None, "lamb", default=lamb) == lamb

    with pytest.raises(ParameterCheckError):
        check.opt_callable_param(2, "lamb")


def test_param_invariant():
    check.param_invariant(True, "some_param")
    num_to_check = 1
    check.param_invariant(num_to_check == 1, "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant(num_to_check == 2, "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant(False, "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant(0, "some_param")

    check.param_invariant(1, "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant("", "some_param")

    check.param_invariant("1kjkjsf", "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant({}, "some_param")

    check.param_invariant({234: "1kjkjsf"}, "some_param")

    with pytest.raises(ParameterCheckError):
        check.param_invariant([], "some_param")

    check.param_invariant([234], "some_param")


def test_string_elem():
    ddict = {"a_str": "a", "a_num": 1, "a_none": None}

    assert check.str_elem(ddict, "a_str") == "a"

    with pytest.raises(ElementCheckError):
        assert check.str_elem(ddict, "a_none")

    with pytest.raises(ElementCheckError):
        check.str_elem(ddict, "a_num")


def test_opt_string_elem():
    ddict = {"a_str": "a", "a_num": 1, "a_none": None}

    assert check.opt_str_elem(ddict, "a_str") == "a"

    assert check.opt_str_elem(ddict, "a_none") == None

    assert check.opt_str_elem(ddict, "nonexistentkey") == None

    with pytest.raises(ElementCheckError):
        check.opt_str_elem(ddict, "a_num")


def test_bool_elem():
    ddict = {"a_true": True, "a_str": "a", "a_num": 1, "a_none": None}

    assert check.bool_elem(ddict, "a_true") is True

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, "a_none")

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, "a_num")

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, "a_str")


def test_invariant():
    assert check.invariant(True)

    with pytest.raises(CheckError):
        check.invariant(False)

    with pytest.raises(CheckError, match="Some Unique String"):
        check.invariant(False, "Some Unique String")

    empty_list = []

    with pytest.raises(CheckError, match="Invariant failed"):
        check.invariant(empty_list)


def test_failed():
    with pytest.raises(CheckError, match="some desc"):
        check.failed("some desc")

    with pytest.raises(CheckError, match="must be a string"):
        check.failed(0)


def test_not_implemented():
    with pytest.raises(NotImplementedCheckError, match="some string"):
        check.not_implemented("some string")

    with pytest.raises(CheckError, match="desc argument must be a string"):
        check.not_implemented(None)


def test_inst():
    class Foo(object):
        pass

    class Bar(object):
        pass

    obj = Foo()

    assert check.inst(obj, Foo) == obj

    with pytest.raises(CheckError, match="not a Bar"):
        check.inst(Foo(), Bar)

    with pytest.raises(CheckError, match="Desc: Expected only a Bar"):
        check.inst(Foo(), Bar, "Expected only a Bar")


def test_inst_param():
    class Foo(object):
        pass

    class Bar(object):
        pass

    class Baaz(object):
        pass

    obj = Foo()

    assert check.inst_param(obj, "obj", Foo) == obj

    with pytest.raises(ParameterCheckError, match="not a Bar"):
        check.inst_param(None, "obj", Bar)

    with pytest.raises(ParameterCheckError, match="not a Bar"):
        check.inst_param(Bar, "obj", Bar)

    with pytest.raises(ParameterCheckError, match="not a Bar"):
        check.inst_param(Foo(), "obj", Bar)

    with pytest.raises(ParameterCheckError, match=r"not one of \['Bar', 'Foo'\]"):
        check.inst_param(None, "obj", (Foo, Bar))

    with pytest.raises(ParameterCheckError, match=r"not one of \['Bar', 'Foo'\]"):
        check.inst_param(Baaz(), "obj", (Foo, Bar))


def test_opt_inst_param():
    class Foo(object):
        pass

    class Bar(object):
        pass

    class Baaz(object):
        pass

    obj = Foo()

    assert check.opt_inst_param(obj, "obj", Foo) == obj
    assert check.opt_inst_param(None, "obj", Foo) is None
    assert check.opt_inst_param(None, "obj", Bar) is None

    with pytest.raises(ParameterCheckError, match="not a Bar"):
        check.opt_inst_param(Bar, "obj", Bar)

    with pytest.raises(ParameterCheckError, match="not a Bar"):
        check.opt_inst_param(Foo(), "obj", Bar)

    # check defaults

    default_obj = Foo()

    assert check.opt_inst_param(None, "obj", Foo, default_obj) is default_obj

    assert check.opt_inst_param(None, "obj", (Foo, Bar)) is None

    with pytest.raises(ParameterCheckError, match=r"not one of \['Bar', 'Foo'\]"):
        check.inst_param(Baaz(), "obj", (Foo, Bar))


def test_dict_elem():
    dict_value = {"blah": "blahblah"}
    ddict = {"dictkey": dict_value, "stringkey": "A", "nonekey": None}

    assert check.dict_elem(ddict, "dictkey") == dict_value

    with pytest.raises(CheckError):
        check.dict_elem(ddict, "stringkey")

    with pytest.raises(CheckError):
        check.dict_elem(ddict, "nonekey")

    with pytest.raises(CheckError):
        check.dict_elem(ddict, "nonexistantkey")


def test_opt_dict_elem():
    dict_value = {"blah": "blahblah"}
    ddict = {"dictkey": dict_value, "stringkey": "A", "nonekey": None}

    assert check.opt_dict_elem(ddict, "dictkey") == dict_value
    assert check.opt_dict_elem(ddict, "nonekey") == {}
    assert check.opt_dict_elem(ddict, "nonexistantkey") == {}

    with pytest.raises(CheckError):
        check.opt_dict_elem(ddict, "stringkey")


def test_list_elem():
    list_value = ["blah", "blahblah"]
    ddict = {"listkey": list_value, "stringkey": "A", "nonekey": None}

    assert check.list_elem(ddict, "listkey") == list_value

    with pytest.raises(CheckError):
        assert check.list_elem(ddict, "nonekey") == []

    with pytest.raises(CheckError):
        assert check.list_elem(ddict, "nonexistantkey") == []

    with pytest.raises(CheckError):
        check.list_elem(ddict, "stringkey")


def test_opt_list_elem():
    list_value = ["blah", "blahblah"]
    ddict = {"listkey": list_value, "stringkey": "A", "nonekey": None}

    assert check.opt_list_elem(ddict, "listkey") == list_value
    assert check.opt_list_elem(ddict, "nonekey") == []
    assert check.opt_list_elem(ddict, "nonexistantkey") == []

    with pytest.raises(CheckError):
        check.opt_list_elem(ddict, "stringkey")


def test_not_none_param():
    assert check.not_none_param(1, "fine")
    check.not_none_param(0, "zero is fine")
    check.not_none_param("", "empty str is fine")

    with pytest.raises(CheckError):
        check.not_none_param(None, "none fails")


def test_is_callable():
    def fn():
        pass

    assert check.is_callable(fn) == fn
    assert check.is_callable(lambda: None)
    assert check.is_callable(lambda: None, "some desc")

    with pytest.raises(CheckError):
        check.is_callable(None)

    with pytest.raises(CheckError):
        check.is_callable(1)

    with pytest.raises(CheckError, match="some other desc"):
        check.is_callable(1, "some other desc")


def test_tuple_param():
    assert check.tuple_param((1, 2), "something")

    with pytest.raises(CheckError):
        assert check.tuple_param(None, "something")

    with pytest.raises(CheckError):
        assert check.tuple_param(1, "something")

    with pytest.raises(CheckError):
        assert check.tuple_param([1], "something")

    with pytest.raises(CheckError):
        assert check.tuple_param({1: 2}, "something")

    with pytest.raises(CheckError):
        assert check.tuple_param("kdjfkd", "something")

    assert check.tuple_param((3, 4), "something", of_type=int)
    assert check.tuple_param(("foo", "bar"), "something", of_type=str)

    assert check.tuple_param((3, 4), "something", of_type=(int, int))
    assert check.tuple_param((3, 4), "something", of_type=(int, int))
    assert check.tuple_param((3, "bar"), "something", of_type=(int, str))

    with pytest.raises(CheckError):
        check.tuple_param((3, 4, 5), "something", of_type=(int, int))

    with pytest.raises(CheckError):
        check.tuple_param((3, 4), "something", of_type=(int, int, int))

    with pytest.raises(CheckError):
        check.tuple_param((3, 4), "something", of_type=(int, str))

    with pytest.raises(CheckError):
        check.tuple_param((3, 4), "something", of_type=(str, str))


def test_matrix_param():
    assert check.matrix_param([[1, 2], [2, 3]], "something")

    with pytest.raises(CheckError):
        assert check.matrix_param(None, "something")

    with pytest.raises(CheckError):
        assert check.matrix_param([1, 2, 4], "something")

    with pytest.raises(CheckError):
        assert check.matrix_param([], "something")

    with pytest.raises(CheckError):
        assert check.matrix_param([[1, 2], 3], "soemthing")

    with pytest.raises(CheckError):
        assert check.matrix_param([[1, 2], [3.0, 4.1]], "something", of_type=int)

    with pytest.raises(CheckError):
        assert check.matrix_param([[1, 2], [2, 3, 4]], "something")


def test_opt_tuple_param():
    assert check.opt_tuple_param((1, 2), "something")
    assert check.opt_tuple_param(None, "something") is None
    assert check.opt_tuple_param(None, "something", (2)) == (2)

    with pytest.raises(CheckError):
        check.opt_tuple_param(1, "something")

    with pytest.raises(CheckError):
        check.opt_tuple_param([1], "something")

    with pytest.raises(CheckError):
        check.opt_tuple_param({1: 2}, "something")

    with pytest.raises(CheckError):
        check.opt_tuple_param("kdjfkd", "something")

    assert check.opt_tuple_param((3, 4), "something", of_type=int)
    assert check.opt_tuple_param(("foo", "bar"), "something", of_type=str)

    assert check.opt_tuple_param((3, 4), "something", of_type=(int, int))
    assert check.opt_tuple_param((3, 4), "something", of_type=(int, int))
    assert check.opt_tuple_param((3, "bar"), "something", of_type=(int, str))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4, 5), "something", of_type=(int, int))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4), "something", of_type=(int, int, int))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4), "something", of_type=(int, str))

    with pytest.raises(CheckError):
        check.opt_tuple_param((3, 4), "something", of_type=(str, str))


def test_opt_type_param():
    class Foo(object):
        pass

    assert check.opt_type_param(int, "foo")
    assert check.opt_type_param(Foo, "foo")

    assert check.opt_type_param(None, "foo") is None
    assert check.opt_type_param(None, "foo", Foo) is Foo

    with pytest.raises(CheckError):
        check.opt_type_param(check, "foo")

    with pytest.raises(CheckError):
        check.opt_type_param(234, "foo")

    with pytest.raises(CheckError):
        check.opt_type_param("bar", "foo")

    with pytest.raises(CheckError):
        check.opt_type_param(Foo(), "foo")


def test_type_param():
    class Bar(object):
        pass

    assert check.type_param(int, "foo")
    assert check.type_param(Bar, "foo")

    with pytest.raises(CheckError):
        check.type_param(None, "foo")

    with pytest.raises(CheckError):
        check.type_param(check, "foo")

    with pytest.raises(CheckError):
        check.type_param(234, "foo")

    with pytest.raises(CheckError):
        check.type_param("bar", "foo")

    with pytest.raises(CheckError):
        check.type_param(Bar(), "foo")


def test_subclass_param():
    class Super(object):
        pass

    class Sub(Super):
        pass

    class Alone(object):
        pass

    assert check.subclass_param(Sub, "foo", Super)

    with pytest.raises(CheckError):
        assert check.subclass_param(Alone, "foo", Super)

    with pytest.raises(CheckError):
        assert check.subclass_param("value", "foo", Super)

    assert check.opt_subclass_param(Sub, "foo", Super)
    assert check.opt_subclass_param(None, "foo", Super) is None

    with pytest.raises(CheckError):
        assert check.opt_subclass_param(Alone, "foo", Super)

    with pytest.raises(CheckError):
        assert check.opt_subclass_param("value", "foo", Super)


@contextmanager
def raises_with_message(exc_type, message_text):
    with pytest.raises(exc_type) as exc_info:
        yield

    assert str(exc_info.value) == message_text


def is_python_three():
    return sys.version_info[0] >= 3


def test_two_dim_dict():
    assert check.two_dim_dict_param({}, "foo") == {}
    assert check.two_dim_dict_param({"key": {}}, "foo")
    assert check.two_dim_dict_param({"key": {"key2": 2}}, "foo")

    # make sure default dict passes
    default_dict = defaultdict(dict)
    default_dict["key"]["key2"] = 2
    assert check.two_dim_dict_param(default_dict, "foo")

    with raises_with_message(
        CheckError,
        """Param "foo" is not a dict. Got None which is type <class 'NoneType'>."""
        if is_python_three()
        else """Param "foo" is not a dict. Got None which is type <type 'NoneType'>.""",
    ):
        check.two_dim_dict_param(None, "foo")

    with raises_with_message(
        CheckError,
        "Value in dictionary mismatches expected type for key int_value. Expected value "
        "of type <class 'dict'>. Got value 2 of type <class 'int'>."
        if is_python_three()
        else "Value in dictionary mismatches expected type for key int_value. Expected value "
        "of type <type 'dict'>. Got value 2 of type <type 'int'>.",
    ):
        check.two_dim_dict_param({"int_value": 2}, "foo")

    with raises_with_message(
        CheckError,
        "Value in dictionary mismatches expected type for key level_two_value_mismatch. "
        "Expected value of type (<class 'str'>,). Got value 2 of type <class 'int'>."
        if is_python_three()
        else "Value in dictionary mismatches expected type for key level_two_value_mismatch. "
        "Expected value of type (<type 'basestring'>,). Got value 2 of type <type 'int'>.",
    ):
        check.two_dim_dict_param(
            {"level_one_key": {"level_two_value_mismatch": 2}}, "foo", value_type=str
        )

    with raises_with_message(
        CheckError,
        "Key in dictionary mismatches type. Expected <class 'int'>. Got 'key'"
        if is_python_three()
        else "Key in dictionary mismatches type. Expected <type 'int'>. Got 'key'",
    ):
        assert check.two_dim_dict_param({"key": {}}, "foo", key_type=int)

    with raises_with_message(
        CheckError,
        "Key in dictionary mismatches type. Expected <class 'int'>. Got 'level_two_key'"
        if is_python_three()
        else "Key in dictionary mismatches type. Expected <type 'int'>. Got 'level_two_key'",
    ):
        assert check.two_dim_dict_param({1: {"level_two_key": "something"}}, "foo", key_type=int)


def test_opt_two_dim_dict_parm():
    assert check.opt_two_dim_dict_param({}, "foo") == {}
    assert check.opt_two_dim_dict_param({"key": {}}, "foo")
    assert check.opt_two_dim_dict_param({"key": {"key2": 2}}, "foo")
    assert check.opt_two_dim_dict_param(None, "foo") == {}

    with pytest.raises(CheckError):
        assert check.opt_two_dim_dict_param("str", "foo")


def test_generator_param():
    def _test_gen():
        yield 1

    assert check.generator_param(_test_gen(), "gen")

    gen = _test_gen()
    assert check.generator(gen)
    assert list(gen) == [1]
    assert check.generator(gen)
    assert list(gen) == []

    with pytest.raises(ParameterCheckError):
        assert check.generator_param(list(gen), "gen")

    with pytest.raises(ParameterCheckError):
        assert check.generator_param(None, "gen")

    with pytest.raises(ParameterCheckError):
        assert check.generator_param(_test_gen, "gen")


def test_opt_generator_param():
    def _test_gen():
        yield 1

    assert check.opt_generator_param(_test_gen(), "gen")

    assert check.opt_generator_param(None, "gen") is None

    with pytest.raises(ParameterCheckError):
        assert check.opt_generator_param(_test_gen, "gen")


def test_generator():
    def _test_gen():
        yield 1

    assert check.generator(_test_gen())

    gen = _test_gen()
    assert check.generator(gen)

    with pytest.raises(ParameterCheckError):
        assert check.generator(list(gen))

    with pytest.raises(ParameterCheckError):
        assert check.generator(None)

    with pytest.raises(ParameterCheckError):
        assert check.generator(_test_gen)


def test_opt_generator():
    def _test_gen():
        yield 1

    assert check.opt_generator(_test_gen())

    gen = _test_gen()
    assert check.opt_generator(gen)
    assert check.opt_generator(None) is None

    with pytest.raises(ParameterCheckError):
        assert check.opt_generator(list(gen))

    with pytest.raises(ParameterCheckError):
        assert check.opt_generator(_test_gen)


def test_internals():
    with pytest.raises(CheckError):
        check._check_key_value_types(None, str, str)  # pylint: disable=protected-access



import pytest

from dagster.core.definitions.utils import (
    check_valid_name,
    config_from_files,
    config_from_pkg_resources,
    config_from_yaml_strings,
)
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from dagster.utils import file_relative_path


def test_check_valid_name():
    assert check_valid_name("a") == "a"

    with pytest.raises(DagsterInvalidDefinitionError):
        assert check_valid_name("has a space")

    with pytest.raises(DagsterInvalidDefinitionError):
        assert check_valid_name("")

    with pytest.raises(DagsterInvalidDefinitionError):
        assert check_valid_name("context")


def test_config_from_files():
    run_config = config_from_files(
        config_files=[file_relative_path(__file__, "./definitions_tests/pass_env.yaml")]
    )

    assert run_config == {"solids": {"can_fail": {"config": {"error": False}}}}

    with pytest.raises(DagsterInvariantViolationError):
        config_from_files(config_files=[file_relative_path(__file__, "not_a_file.yaml")])

    with pytest.raises(DagsterInvariantViolationError):
        config_from_files(
            config_files=[
                file_relative_path(__file__, "./definitions_tests/test_repository_definition.py")
            ]
        )


def test_config_from_yaml_strings():
    a = """
foo:
  bar: 1
baz: 3
"""
    b = """
foo:
  one: "one"
other: 4
"""
    c = """
final: "result"
"""

    run_config = config_from_yaml_strings([a, b, c])
    assert run_config == {
        "foo": {"bar": 1, "one": "one"},
        "baz": 3,
        "other": 4,
        "final": "result",
    }

    with pytest.raises(
        DagsterInvariantViolationError, match="Encountered error attempting to parse yaml"
    ):
        config_from_yaml_strings(["--- `"])

    run_config = config_from_yaml_strings([])
    assert run_config == {}


def test_config_from_pkg_resources():
    good = (
        "dagster_tests.core_tests.definitions_tests",
        "pass_env.yaml",
    )
    run_config = config_from_pkg_resources([good])
    assert run_config == {"solids": {"can_fail": {"config": {"error": False}}}}

    bad_defs = [
        ("dagster_tests.core_tests.definitions_tests", "does_not_exist.yaml"),
        ("dagster_tests.core_tests.definitions_tests", "bad_file_binary.yaml",),
    ]

    for bad_def in bad_defs:
        with pytest.raises(
            DagsterInvariantViolationError, match="Encountered error attempting to parse yaml",
        ):
            config_from_pkg_resources([bad_def])



import pytest

from dagster import (
    DagsterInvalidDefinitionError,
    InputDefinition,
    Nothing,
    Optional,
    composite_solid,
    execute_pipeline,
    execute_solid,
    lambda_solid,
    pipeline,
)


def test_none():
    @lambda_solid(input_defs=[InputDefinition("x", Optional[int], default_value=None)])
    def none_x(x):
        return x

    result = execute_solid(none_x)
    assert result.output_value() == None


def test_none_infer():
    @lambda_solid
    def none_x(x=None):
        return x

    result = execute_solid(none_x)
    assert result.output_value() == None


def test_int():
    @lambda_solid(input_defs=[InputDefinition("x", Optional[int], default_value=1337)])
    def int_x(x):
        return x

    result = execute_solid(int_x)
    assert result.output_value() == 1337


def test_int_infer():
    @lambda_solid
    def int_x(x=1337):
        return x

    result = execute_solid(int_x)
    assert result.output_value() == 1337


def test_early_fail():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match="Type check failed for the default_value of InputDefinition x of type Int",
    ):

        @lambda_solid(input_defs=[InputDefinition("x", int, default_value="foo")])
        def _int_x(x):
            return x

    with pytest.raises(
        DagsterInvalidDefinitionError,
        match="Type check failed for the default_value of InputDefinition x of type String",
    ):

        @lambda_solid(input_defs=[InputDefinition("x", str, default_value=1337)])
        def _int_x(x):
            return x


# we can't catch bad default_values except for scalars until runtime since the type_check function depends on
# a context that has access to resources etc.
@lambda_solid(input_defs=[InputDefinition("x", Optional[int], default_value="number")])
def bad_default(x):
    return x


def test_mismatch():
    result = execute_solid(bad_default, raise_on_error=False)
    assert result.success == False
    assert result.input_events_during_compute[0].step_input_data.type_check_data.success == False


def test_env_precedence():
    result = execute_solid(
        bad_default,
        run_config={"solids": {"bad_default": {"inputs": {"x": 1}}}},
        raise_on_error=False,
    )
    assert result.success == True
    assert result.output_value() == 1


def test_input_precedence():
    @lambda_solid
    def emit_one():
        return 1

    @pipeline
    def pipe():
        bad_default(emit_one())

    result = execute_pipeline(pipe)
    assert result.success
    assert result.output_for_solid("bad_default") == 1


def test_nothing():
    with pytest.raises(DagsterInvalidDefinitionError):

        @lambda_solid(input_defs=[InputDefinition("x", Nothing, default_value=None)])
        def _nothing():
            pass


def test_composite_outer_default():
    @lambda_solid(input_defs=[InputDefinition("x", Optional[int])])
    def int_x(x):
        return x

    @composite_solid(input_defs=[InputDefinition("y", Optional[int], default_value=42)])
    def wrap(y):
        return int_x(y)

    result = execute_solid(wrap)
    assert result.success
    assert result.output_value() == 42


def test_composite_inner_default():
    @lambda_solid(input_defs=[InputDefinition("x", Optional[int], default_value=1337)])
    def int_x(x):
        return x

    @composite_solid(input_defs=[InputDefinition("y", Optional[int])])
    def wrap(y):
        return int_x(y)

    result = execute_solid(wrap)
    assert result.success
    assert result.output_value() == 1337


def test_composite_precedence_default():
    @lambda_solid(input_defs=[InputDefinition("x", Optional[int], default_value=1337)])
    def int_x(x):
        return x

    @composite_solid(input_defs=[InputDefinition("y", Optional[int], default_value=42)])
    def wrap(y):
        return int_x(y)

    result = execute_solid(wrap)
    assert result.success
    assert result.output_value() == 42


def test_composite_mid_default():
    @lambda_solid(input_defs=[InputDefinition("x", Optional[int])])
    def int_x(x):
        return x

    @composite_solid(input_defs=[InputDefinition("y", Optional[int], default_value=42)])
    def wrap(y):
        return int_x(y)

    @composite_solid(input_defs=[InputDefinition("z", Optional[int])])
    def outter_wrap(z):
        return wrap(z)

    result = execute_solid(outter_wrap)
    assert result.success
    assert result.output_value() == 42



from functools import update_wrapper

from dagster import check
from dagster.config.field_utils import check_user_facing_opt_config_param
from dagster.core.definitions.config_mappable import IConfigMappable
from dagster.utils.backcompat import rename_warning


class IntermediateStorageDefinition(IConfigMappable):
    """Defines intermediate data storage behaviors.
    Args:
        name (str): Name of the storage mode.
        is_persistent (bool): Whether the storage is persistent in a way that can cross process/node
            boundaries. Re-execution with, for example, the multiprocess executor, or with
            dagster-airflow, requires a persistent storage mode.
        required_resource_keys(Optional[Set[str]]): The resources that this storage needs at runtime to function.
        config_schema (Optional[ConfigSchema]): The schema for the storage's configuration schema.
            Configuration data passed in this schema will be made available to the
            ``intermediate_storage_creation_fn`` under ``init_context.intermediate_storage_config``.
        intermediate_storage_creation_fn: (Callable[[InitIntermediateStorageContext], IntermediateStorage])
            Called to construct the storage. This function should consume the init context and emit
            a :py:class:`IntermediateStorage`.
        _configured_config_mapping_fn: This argument is for internal use only. Users should not
            specify this field. To preconfigure a resource, use the :py:func:`configured` API.
    """

    def __init__(
        self,
        name,
        is_persistent,
        required_resource_keys,
        config_schema=None,
        intermediate_storage_creation_fn=None,
        _configured_config_mapping_fn=None,
    ):
        self._name = check.str_param(name, "name")
        self._is_persistent = check.bool_param(is_persistent, "is_persistent")
        self._config_schema = check_user_facing_opt_config_param(config_schema, "config_schema")
        self._intermediate_storage_creation_fn = check.opt_callable_param(
            intermediate_storage_creation_fn, "intermediate_storage_creation_fn"
        )
        self._required_resource_keys = frozenset(
            check.set_param(
                required_resource_keys if required_resource_keys else set(),
                "required_resource_keys",
                of_type=str,
            )
        )
        self.__configured_config_mapping_fn = check.opt_callable_param(
            _configured_config_mapping_fn, "config_mapping_fn"
        )

    @property
    def config_field(self):
        rename_warning("config_schema", "config_field", "0.9.0")
        return self.config_schema

    @property
    def name(self):
        return self._name

    @property
    def is_persistent(self):
        return self._is_persistent

    @property
    def config_schema(self):
        return self._config_schema

    @property
    def intermediate_storage_creation_fn(self):
        return self._intermediate_storage_creation_fn

    @property
    def required_resource_keys(self):
        return self._required_resource_keys

    @property
    def _configured_config_mapping_fn(self):
        return self.__configured_config_mapping_fn

    def configured(self, config_or_config_fn, config_schema=None, **kwargs):
        """
        Wraps this object in an object of the same type that provides configuration to the inner
        object.

        Args:
            config_or_config_fn (Union[Any, Callable[[Any], Any]]): Either (1) Run configuration
                that fully satisfies this object's config schema or (2) A function that accepts run
                configuration and returns run configuration that fully satisfies this object's
                config schema.  In the latter case, config_schema must be specified.  When
                passing a function, it's easiest to use :py:func:`configured`.
            config_schema (ConfigSchema): If config_or_config_fn is a function, the config schema
                that its input must satisfy.
            name (Optional[str]): Name of the storage mode. If not specified, inherits the name
                of the storage mode being configured.

        Returns (IntermediateStorageDefinition): A configured version of this object.
        """
        name = check.opt_str_param(kwargs.get("name"), "name", self.name)
        wrapped_config_mapping_fn = self._get_wrapped_config_mapping_fn(
            config_or_config_fn, config_schema
        )

        return IntermediateStorageDefinition(
            name=name,
            is_persistent=self.is_persistent,
            required_resource_keys=self.required_resource_keys,
            config_schema=config_schema,
            intermediate_storage_creation_fn=self.intermediate_storage_creation_fn,
            _configured_config_mapping_fn=wrapped_config_mapping_fn,
        )


def intermediate_storage(
    required_resource_keys=None, name=None, is_persistent=True, config_schema=None
):
    """Creates an intermediate storage definition

    The decorated function will be passed as the ``intermediate_storage_creation_fn`` to a
    :py:class:`IntermediateStorageDefinition`.

    Args:
        name (str): The name of the intermediate storage.
        is_persistent (bool): Whether the storage is persistent in a way that can cross process/node
            boundaries. Re-execution with, for example, the multiprocess executor, or with
            dagster-airflow, requires a persistent storage mode.
        required_resource_keys (Optional[Set[str]]):
            The resources that this storage needs at runtime to function.
        config_schema (Optional[ConfigSchema]): The schema for the config. Configuration data available in
            `init_context.intermediate_storage_config`.
    """

    if callable(name):
        check.invariant(is_persistent is True)
        check.invariant(config_schema is None)
        check.invariant(required_resource_keys is None)
        return _IntermediateStorageDecoratorCallable()(name)

    return _IntermediateStorageDecoratorCallable(
        name=name,
        is_persistent=is_persistent,
        config_schema=config_schema,
        required_resource_keys=required_resource_keys,
    )


class _IntermediateStorageDecoratorCallable(object):
    def __init__(
        self, name=None, is_persistent=True, config_schema=None, required_resource_keys=None
    ):
        self.name = check.opt_str_param(name, "name")
        self.is_persistent = check.bool_param(is_persistent, "is_persistent")
        self.config_schema = check_user_facing_opt_config_param(config_schema, "config_schema")
        self.required_resource_keys = check.opt_set_param(
            required_resource_keys, "required_resource_keys", of_type=str
        )

    def __call__(self, fn):
        check.callable_param(fn, "fn")

        if not self.name:
            self.name = fn.__name__

        storage_def = IntermediateStorageDefinition(
            name=self.name,
            is_persistent=self.is_persistent,
            config_schema=self.config_schema,
            intermediate_storage_creation_fn=fn,
            required_resource_keys=self.required_resource_keys,
        )

        update_wrapper(storage_def, wrapped=fn)

        return storage_def
